
__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


import click

from akit.environment.variables import LOG_LEVEL_NAMES

from akit.cli.cmdtree.generators.upnp import group_akit_generators_upnp

@click.group("generators", help="Contains code generators for creating various types of interop code.")
def group_akit_generators():
    return

group_akit_generators.add_command(group_akit_generators_upnp)



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class CallManagement1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:CallManagement:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'CallManagement1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "CallBackAvailability": { "data_type": "string", "default": None, "allowed_list": None},
        "CallInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "TelCPNameList": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AcceptCall(self, TelCPName, SecretKey, TargetCallID, MediaCapabilityInfo, CallMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AcceptCall action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
            "MediaCapabilityInfo": MediaCapabilityInfo,
            "CallMode": CallMode,
        }

        self.call_action("AcceptCall", arguments=arguments, aspects=aspects)

        return

    def action_AcceptModifyCall(self, TelCPName, SecretKey, TargetCallID, MediaCapabilityInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AcceptModifyCall action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
            "MediaCapabilityInfo": MediaCapabilityInfo,
        }

        self.call_action("AcceptModifyCall", arguments=arguments, aspects=aspects)

        return

    def action_ChangeMonopolizer(self, CurrentMonopolizer, SecretKey, CallID, NewMonopolizer, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ChangeMonopolizer action.
        """
        arguments = {
            "CurrentMonopolizer": CurrentMonopolizer,
            "SecretKey": SecretKey,
            "CallID": CallID,
            "NewMonopolizer": NewMonopolizer,
        }

        self.call_action("ChangeMonopolizer", arguments=arguments, aspects=aspects)

        return

    def action_ChangeTelCPName(self, CurrentTelCPName, CurrentSecretKey, NewTelCPName, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ChangeTelCPName action.

            :returns: "NewSecretKey", "Expires"
        """
        arguments = {
            "CurrentTelCPName": CurrentTelCPName,
            "CurrentSecretKey": CurrentSecretKey,
            "NewTelCPName": NewTelCPName,
        }

        out_params = self.call_action("ChangeTelCPName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSecretKey", "Expires",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ClearCallBack(self, CallBackID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ClearCallBack action.
        """
        arguments = {
            "CallBackID": CallBackID,
        }

        self.call_action("ClearCallBack", arguments=arguments, aspects=aspects)

        return

    def action_ClearCallLogs(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ClearCallLogs action.
        """
        arguments = { }

        self.call_action("ClearCallLogs", arguments=arguments, aspects=aspects)

        return

    def action_GetCallBackInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCallBackInfo action.

            :returns: "CallBackInfo"
        """
        arguments = { }

        out_params = self.call_action("GetCallBackInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallBackInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCallInfo(self, TelCPName, SecretKey, TargetCallID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCallInfo action.

            :returns: "CallInfoList"
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
        }

        out_params = self.call_action("GetCallInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallInfoList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCallLogs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCallLogs action.

            :returns: "CallLogs"
        """
        arguments = { }

        out_params = self.call_action("GetCallLogs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallLogs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaCapabilities(self, TCMediaCapabilityInfo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaCapabilities action.

            :returns: "SupportedMediaCapabilityInfo"
        """
        arguments = {
            "TCMediaCapabilityInfo": TCMediaCapabilityInfo,
        }

        out_params = self.call_action("GetMediaCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedMediaCapabilityInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTelCPNameList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTelCPNameList action.

            :returns: "TelCPNameList"
        """
        arguments = { }

        out_params = self.call_action("GetTelCPNameList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TelCPNameList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTelephonyIdentity(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTelephonyIdentity action.

            :returns: "TelephonyIdentity"
        """
        arguments = { }

        out_params = self.call_action("GetTelephonyIdentity", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TelephonyIdentity",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_InitiateCall(self, CalleeID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the InitiateCall action.

            :returns: "CallID"
        """
        arguments = {
            "CalleeID": CalleeID,
        }

        out_params = self.call_action("InitiateCall", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ModifyCall(self, TelCPName, SecretKey, TargetCallID, MediaCapabilityInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ModifyCall action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
            "MediaCapabilityInfo": MediaCapabilityInfo,
        }

        self.call_action("ModifyCall", arguments=arguments, aspects=aspects)

        return

    def action_RegisterCallBack(self, CalleeID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterCallBack action.

            :returns: "CallBackID"
        """
        arguments = {
            "CalleeID": CalleeID,
        }

        out_params = self.call_action("RegisterCallBack", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallBackID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RegisterTelCPName(self, TelCPName, CurrentSecretKey, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterTelCPName action.

            :returns: "NewSecretKey", "Expires"
        """
        arguments = {
            "TelCPName": TelCPName,
            "CurrentSecretKey": CurrentSecretKey,
        }

        out_params = self.call_action("RegisterTelCPName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSecretKey", "Expires",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RejectCall(self, TelCPName, SecretKey, TargetCallID, RejectReason, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RejectCall action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
            "RejectReason": RejectReason,
        }

        self.call_action("RejectCall", arguments=arguments, aspects=aspects)

        return

    def action_StartCall(self, TelCPName, SecretKey, CalleeID, CallPriority, MediaCapabilityInfo, CallMode, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartCall action.

            :returns: "CallID"
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "CalleeID": CalleeID,
            "CallPriority": CallPriority,
            "MediaCapabilityInfo": MediaCapabilityInfo,
            "CallMode": CallMode,
        }

        out_params = self.call_action("StartCall", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CallID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_StartMediaTransfer(self, TelCPName, SecretKey, TargetCallID, TCList, MediaCapabilityInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartMediaTransfer action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "TargetCallID": TargetCallID,
            "TCList": TCList,
            "MediaCapabilityInfo": MediaCapabilityInfo,
        }

        self.call_action("StartMediaTransfer", arguments=arguments, aspects=aspects)

        return

    def action_StopCall(self, TelCPName, SecretKey, CallID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StopCall action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
            "CallID": CallID,
        }

        self.call_action("StopCall", arguments=arguments, aspects=aspects)

        return

    def action_UnregisterTelCPName(self, TelCPName, SecretKey, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UnregisterTelCPName action.
        """
        arguments = {
            "TelCPName": TelCPName,
            "SecretKey": SecretKey,
        }

        self.call_action("UnregisterTelCPName", arguments=arguments, aspects=aspects)

        return



"""
.. module:: landscapeconfigurationlayer
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the Landscape related classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import List, Optional, Union

import os
import pprint
import shutil
import threading

from akit.exceptions import AKitConfigurationError
from akit.friendlyidentifier import FriendlyIdentifier
from akit.paths import get_filename_for_landscape, get_filename_for_runtime, get_filename_for_topology, get_path_for_output

from akit.xformatting import split_and_indent_lines

from akit.interop.credentials.credentialmanager import CredentialManager

from akit.xfeature import FeatureFilter, FeatureTag, FeatureMask
from akit.interop.landscaping.landscapedescription import LandscapeDescription
from akit.interop.landscaping.landscapedevice import LandscapeDevice
from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension
from akit.interop.landscaping.topologydescription import TopologyDescription

PASSWORD_MASK = "(hidden)"

def mask_passwords (context):
    """
        Takes a dictionary context object and will recursively mask any password members found
        in the dictionary.
    """
    for key, val in context.items():
        if (key == "password" or key == "secret"):
            context[key] = PASSWORD_MASK

        if isinstance(val, dict):
            mask_passwords(val)
        elif isinstance(val, list):
            for item in val:
                if isinstance(item, dict):
                    mask_passwords(item)

    return

def filter_credentials(device_info, credential_lookup, category):
    """
        Looks up the credentials associated with a device and returns the credentials found
        that match a given category.

        :param device_info: Device information dictionary with credential names to reference.
        :param credential_lookup: A credential lookup dictionary that is used to convert credential
                                  names into credential objects loaded from the landscape.
        :param category: The category of credentials to return when filtering credentials.
    """
    cred_found_list = []

    cred_name_list = device_info["credentials"]
    for cred_name in cred_name_list:
        if cred_name in credential_lookup:
            credential = credential_lookup[cred_name]
            if credential.category == category:
                cred_found_list.append(credential)
        else:
            error_lines = [
                "The credential '{}' was not found in the credentials list.",
                "DEVICE:"
            ]

            dev_repr_lines = pprint.pformat(device_info, indent=4).splitlines(False)
            for dline in dev_repr_lines:
                error_lines.append("    " + dline)

            error_lines.append("CREDENTIALS:")
            cred_available_list = [cname for cname in credential_lookup.keys()]
            cred_available_list.sort()
            for cred_avail in cred_available_list:
                error_lines.append("    " + cred_avail)

            errmsg = os.linesep.join(error_lines)
            raise AKitConfigurationError(errmsg) from None

    return cred_found_list

# ====================================================================================
#
#                                     CONFIGURATION LAYER
#
# ====================================================================================
class LandscapeConfigurationLayer:
    """
        The :class:`LandscapeConfigurationLayer` serves as the base layer for the :class:`Landscape` object.  The
        :class:`LandscapeConfigurationLayer` contains the data and method that are initilized as part of the
        initialization of the Landscape object.  It allows access to the processed data pulled from the
        "landscape.yaml" file which details the static declarations for the devices and resources that
        are the landscape file declares.
    """

    landscape_description = LandscapeDescription
    landscape_device = LandscapeDevice
    landscape_device_extension = LandscapeDeviceExtension

    topology_description = TopologyDescription

    _configured_gate = threading.Event()
    _configuring_thread_id = None

    def __init__(self):
        """
            The :class:`LandscapeConfigurationLayer` object should not be instantiated directly.
        """
        self._landscape_info = None
        self._landscape_file = None

        self._topology_info = None
        self._topology_file = None

        self._environment_info = None
        self._environment_label = None
     
        self._runtime_info = None

        self._has_upnp_devices = False
        self._has_ssh_devices = False

        self._all_devices = {}

        self._credentials = {}

        self._serial_config_lookup_table = {}

        super().__init__()

        return

    @property
    def databases(self) -> dict:
        """
            Returns the database configuration information from the landscape file.
        """
        db_info = self._runtime_info["databases"]
        return db_info

    @property
    def credentials(self) -> dict:
        return self._credentials

    @property
    def environment(self) -> dict:
        """
            Returns the environment section of the landscape configuration.
        """
        return self._environment_info

    @property
    def environment_label(self) -> str:
        """
            Returns the environment.label section of the landscape configuration.
        """
        return self._environment_label

    @property
    def landscape_info(self):
        """
            Returns the root landscape configuration dictionary.
        """
        return self._landscape_info

    @property
    def has_ssh_devices(self) -> bool:
        """
            Returns a boolean indicating if the landscape contains ssh devices.
        """
        return self._has_ssh_devices

    @property
    def has_upnp_devices(self) -> bool:
        """
            Returns a boolean indicating if the landscape contains upnp devices.
        """
        return self._has_upnp_devices

    @property
    def name(self) -> str:
        """
            Returns the name associated with the landscape.
        """
        lname = None
        if "name" in self.landscape_info:
            lname = self.landscape_info["name"]
        return lname

    @property
    def networking(self) -> dict:
        """
            Returns the configuration/networking section of the runtime configuration.
        """
        netinfo = None
        if "networking" in self._runtime_info:
            netinfo = self._runtime_info["networking"]
        return netinfo

    def activate_configuration(self):
        """
            Called once at the beginning of the lifetime of a Landscape derived type.  This
            allows the derived types to participate in a customized intialization process.
        """

        lscapeType = type(self)

        lscapeType.landscape_lock.acquire()
        try:

            if lscapeType._configuring_thread_id is None:
                init_thread = threading.current_thread()
                lscapeType._configuring_thread_id = init_thread.ident

                # We don't need to hold the landscape lock while initializing
                # the Landscape because no threads calling the constructor can
                # exit without the landscape initialization being finished.
                lscapeType.landscape_lock.release()

                try:
                    log_to_directory = None

                    log_landscape_declaration = lscapeType.context.lookup("/environment/behaviors/log-landscape-declaration")
                    if log_landscape_declaration:
                        log_to_directory = get_path_for_output()

                    self._load_runtime(log_to_directory=log_to_directory)

                    self._load_landscape(log_to_directory=log_to_directory)

                    self._load_topology(log_to_directory=log_to_directory)

                    self._initialize_landscape()

                    # Set the landscape_initialized even to allow other threads to use the APIs of the Landscape object
                    self._configured_gate.set()

                finally:
                    lscapeType.landscape_lock.acquire()

            else:

                # Don't hold the landscape like while we wait for the
                # landscape to be initialized
                lscapeType.landscape_lock.release()
                try:
                    # Because the landscape is a global singleton and because
                    # we were not the first thread to call the contructor, wait
                    # for the first calling thread to finish initializing the
                    # Landscape before we return and try to use the returned
                    # Landscape reference
                    self._configured_gate.wait()
                finally:
                    lscapeType.landscape_lock.acquire()

        finally:
            lscapeType.landscape_lock.release()

        return

    def get_device_configs(self) -> List[dict]:
        """
            Returns the list of device configurations from the landscape.  This will
            skip any device that has a "skip": true member.
        """
        device_config_list = self._internal_get_device_configs()

        return device_config_list

    def get_devices(self) -> List[LandscapeDevice]:
        """
            Returns the list of devices from the landscape.  This will
            skip any device that has a "skip": true member.
        """
        device_list = None

        self.landscape_lock.acquire()
        try:
            device_list = [dev for dev in self._all_devices.values()]
        finally:
            self.landscape_lock.release()

        return device_list

    def get_devices_with_feature(self, feature: Union[FeatureTag, str]) -> List[LandscapeDevice]:
        """
            Returns the list of devices from the landscape.  This will
            skip any device that has a "skip": true member.
        """
        device_list = None

        if isinstance(feature, FeatureTag):
            feature = feature.ID

        self.landscape_lock.acquire()
        try:
            device_list = []
            for dev in self._all_devices.values():
                if dev.has_feature(feature):
                    device_list.append(dev)
        finally:
            self.landscape_lock.release()

        return device_list

    def get_devices_with_feature_mask(self, feature_mask: FeatureMask) -> List[LandscapeDevice]:
        """
            Returns the list of devices from the landscape.  This will
            skip any device that has a "skip": true member.
        """

        all_devices = self.get_devices()

        feature_filter = FeatureFilter(**feature_mask)

        filtered_devices = feature_filter.filter()

        return filtered_devices

    def get_ssh_device_configs(self, exclude_upnp=False) -> List[dict]:
        """
            Returns a list of devices that support ssh.
        """
        ssh_device_config_list = []

        for devinfo in self._internal_get_device_configs():
            dev_type = devinfo["deviceType"]

            if exclude_upnp and dev_type == "network/upnp":
                continue

            if dev_type == "network/ssh":
                ssh_device_config_list.append(devinfo)

        return ssh_device_config_list

    def get_ssh_device_list(self) -> List[dict]:
        """
            Returns a list of SSH devices.
        """

        ssh_device_list = []

        for device in self.get_devices():
            device_type = device.device_type
            if device_type == "network/ssh":
                ssh_device_list.append(device)
            elif device_type == "network/upnp":
                if device.has_ssh_credential:
                    ssh_device_list.append(device)

        return ssh_device_list

    def get_upnp_device_configs(self, ssh_only=False) -> List[dict]:
        """
            Returns a list of UPNP device information dictionaries.
        """
        upnp_device_config_list = self._internal_get_upnp_device_configs(ssh_only=ssh_only)

        return upnp_device_config_list

    def get_upnp_device_config_lookup_table(self) -> dict:
        """
            Returns a USN lookup table for upnp devices.
        """
        upnp_device_table = self._internal_get_upnp_device_config_lookup_table()

        return upnp_device_table

    def get_upnp_device_list(self) -> List[dict]:
        """
            Returns a list of UPNP devices.
        """

        upnp_device_list = []

        for device in self.get_devices():
            device_type = device.device_type
            if device_type == "network/upnp":
                upnp_device_list.append(device)

        return upnp_device_list

    def get_serial_config(self, serial_service_name: str):
        """
            Looks up the configuration dictionary for the serial service specified.
        """
        serial_config = None

        pod_config = self._landscape_info["pod"]

        if "serial" in pod_config:
            if self._serial_config_lookup_table is not None:
                serial_config_lookup_table = self._serial_config_lookup_table
            else:
                serial_config_lookup_table = {}

                serial_config_list = pod_config["serial"]
                for serial_config in serial_config_list:
                    cfgname = serial_config["name"]
                    serial_config_lookup_table[cfgname] = serial_config

            if serial_service_name in self._serial_config_lookup_table:
                serial_config = self._serial_config_lookup_table[serial_service_name]

        return serial_config

    def _create_landscape_device(self, friendly_id: FriendlyIdentifier, dev_type: str, dev_config_info: dict) -> LandscapeDevice:
        device = None

        self.landscape_lock.acquire()
        try:
            identity = friendly_id.identity
            if identity in self._all_devices:
                device = self._all_devices[identity]
            else:
                lscape = self
                device = LandscapeDevice(lscape, friendly_id, dev_type, dev_config_info)
                self._all_devices[identity] = device
        finally:
            self.landscape_lock.release()

        return device

    def _enhance_landscape_device(self, landscape_device, primary_dev_extension):
        return landscape_device

    def _initialize_credentials(self):
        """
        """
        credmgr = CredentialManager()

        self._credentials = credmgr.credentials

        return

    def _initialize_devices(self):

        for dev_config_info in self._internal_get_device_configs():
            dev_type = dev_config_info["deviceType"]
            
            self._initialize_device_of_type(dev_type, dev_config_info)

        return

    def _initialize_device_of_type(self, dev_type: str, dev_config_info: dict):

        if dev_type == "network/upnp":
            upnp_info = dev_config_info["upnp"]

            devhint = None
            if "hint" in upnp_info:
                devhint = upnp_info["hint"]
            elif "USN" in upnp_info:
                devhint = upnp_info["USN"]
                if "hint" not in upnp_info:
                    upnp_info["hint"] = devhint

            fdid = FriendlyIdentifier(devhint, devhint)
            self._create_landscape_device(fdid, dev_type, dev_config_info)
        elif dev_type == "network/ssh":
            devhint = dev_config_info["host"]
            fdid = FriendlyIdentifier(devhint, devhint)
            self._create_landscape_device(fdid, dev_type, dev_config_info)
        else:
            errmsg_lines = [
                "Unknown device type %r in configuration file." % dev_type,
                "DEVICE INFO:"
            ]
            errmsg_lines.extend(split_and_indent_lines(pprint.pformat(dev_config_info, indent=4), 1))

            errmsg = os.linesep.join(errmsg_lines)
            raise AKitConfigurationError(errmsg) from None

        return

    def _initialize_landscape(self):

        self._initialize_credentials()

        # Initialize the devices so we know what they are, this will create a LandscapeDevice object for each device
        # and register it in the all_devices table where it can be found by the device coordinators for further activation
        self._initialize_devices()

        self._topology_overlay()

        return

    def _internal_get_device_configs(self) -> List[dict]:
        """
            Returns the list of devices from the landscape.  This will
            skip any device that has a "skip": true member.

            .. note:: The _internal_ methods do not guard against calls prior to
            landscape initialization so they should only be called with care.  This
            should not be called until after the _landscape_info variable has been
            loaded and contains the configuration data from the landscape.yaml file.
        """

        device_config_list = []

        self.landscape_lock.acquire()
        try:
            pod_info = self._landscape_info["pod"]
            for dev_config_info in pod_info["devices"]:
                if "skip" in dev_config_info and dev_config_info["skip"]:
                    continue
                device_config_list.append(dev_config_info)
        finally:
            self.landscape_lock.release()

        return device_config_list

    def _internal_get_ssh_device_configs(self) -> List[dict]:
        """
            Returns a list of SSH device information dictionaries.
        """

        ssh_device_config_list = []

        for device_config in self._internal_get_device_configs():
            dev_type = device_config["deviceType"]

            if dev_type == "network/ssh":
                ssh_device_config_list.append(device_config)

        return ssh_device_config_list

    def _internal_get_upnp_device_configs(self, ssh_only=False) -> List[dict]:
        """
            Returns a list of UPNP device information dictionaries.
        """

        upnp_device_config_list = []

        for device_config in self._internal_get_device_configs():
            dev_type = device_config["deviceType"]

            if dev_type != "network/upnp":
                continue

            if ssh_only and "ssh" in device_config:
                upnp_device_config_list.append(device_config)
            else:
                upnp_device_config_list.append(device_config)

        return upnp_device_config_list

    def _internal_get_upnp_device_list(self) -> List[dict]:
        """
            Returns a list of UPNP devices.
        """

        upnp_device_list = []

        for device in self._all_devices.values():
            if device.device_type == "network/upnp":
                upnp_device_list.append(device)

        return upnp_device_list

    
    def _internal_get_upnp_device_config_lookup_table(self) -> dict:
        """
            Returns a USN lookup table for upnp devices.

            .. note:: The _internal_ methods do not guard against calls prior to
            landscape initialization so they should only be called with care.  This
            should not be called until after the _landscape_info variable has been
            loaded and contains the configuration data from the landscape.yaml file.
        """

        upnp_device_config_list = self._internal_get_upnp_device_configs()

        upnp_device_config_table = {}
        for device_config in upnp_device_config_list:
            upnp_info = device_config["upnp"]

            dkey = None
            if "hint" in upnp_info:
                dkey = upnp_info["hint"]
            elif "USN" in upnp_info:
                dkey = upnp_info["USN"]
            else:
                errmsg = "The UPNP configuration error should contain either a 'hint' or 'USN' field."
                raise AKitConfigurationError(errmsg)

            upnp_device_config_table[dkey] = device_config

        return upnp_device_config_table

    def _internal_lookup_device_by_identity(self, identity: str) -> Optional[LandscapeDevice]:
        """
            Looks up a device by identity.
        """

        self.landscape_lock.acquire()
        try:
            device = None
            if identity in self._all_devices:
                device = self._all_devices[identity]
        finally:
            self.landscape_lock.release()

        return device
    
    def _internal_lookup_device_by_hint(self, hint: str) -> Optional[LandscapeDevice]:
        """
            Looks up a device by identity.
        """

        self.landscape_lock.acquire()
        try:
            device = None
            for identity, dev in self._all_devices.items():
                if identity.find(hint) > -1:
                    device = dev
                    break
        finally:
            self.landscape_lock.release()

        return device

    def _load_landscape(self, log_to_directory: Optional[str]=None):

        self._landscape_file = get_filename_for_landscape()

        landscape_desc = self.landscape_description()
        self._landscape_info = landscape_desc.load(self._landscape_file, log_to_directory=log_to_directory)

        if "environment" not in self._landscape_info:
            err_msg = "The landscape file must have an 'environment' decription. (%s)" % self._landscape_file
            raise AKitConfigurationError(err_msg) from None

        self._environment_info = self._landscape_info["environment"]
        if "label" not in self._environment_info:
            err_msg = "The landscape 'environment' decription must have a 'label' member (development, production, test). (%s)" % self._landscape_file
            raise AKitConfigurationError(err_msg) from None

        return

    def _load_runtime(self, log_to_directory: Optional[str]=None):

        lscapeType = type(self)

        if log_to_directory is not None:
            # Log the runtime file used to startup the landscape configuration
            runtime_file = get_filename_for_runtime()
            if os.path.exists(runtime_file):
                runtime_basename = os.path.basename(runtime_file)
                runtime_logged_file = os.path.join(log_to_directory, runtime_basename)

                shutil.copy(runtime_file, runtime_logged_file)

        self._runtime_info = lscapeType.context.lookup("/configuration")

        return

    def _load_topology(self, log_to_directory: Optional[str]=None):
        """
            Loads the topology file.
        """

        self._topology_file = get_filename_for_topology()

        topology_desc = self.topology_description()
        self._topology_info = topology_desc.load(self._topology_file, log_to_directory=log_to_directory)

        return

    def _topology_overlay(self):
        return

    def _topology_validate_specification(self, topology_info):
        return


"""
.. module:: coordinatorbase
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the CoordinatorBase which is the base object for coordinators to derive from and establishes
               patterns for coordinators which help to make them threadsafe.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, TYPE_CHECKING

import threading
import weakref

from akit.exceptions import AKitNotOverloadedError
from akit.interop.landscaping.landscapedevice import LandscapeDevice
from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension

from akit.xlogging.foundations import getAutomatonKitLogger

EMPTY_LIST = []

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class CoordinatorBase:
    """
        The CoordinatorBase utilizes the expected device declarations of type such as 'network/upnp' to establish and maintain
        connectivity and interoperability with a class of devices.  A derived coordinator will scan the medium such as a network
        for the devices declared in the landscape description.  The coordinator will also create the threads necessary to maintain
        communicates with the external devices over the medium.
    """

    instance = None
    initialized = False

    def __new__(cls, *_args, **_kwargs):
        """
            Constructs new instances of the :class:`UpnpCoordinator` object. The
            :class:`UpnpCoordinator` object is a singleton so following instantiations
            of the object will reference the existing singleton
        """

        if cls.instance is None:
            cls.instance = super(CoordinatorBase, cls).__new__(cls)
        return cls.instance

    def __init__(self, lscape: "Landscape", *args, coord_config=None, **kwargs):
        """
            Constructs an instance of a derived :class:`CoordinatorBase` object.

            :param lscape: The :class:`Landscape` singleton instance.
            :param *args: A pass through for other positional args.
            :param **kwargs: A pass through for the other keyword args.
        """
        this_type = type(self)
        if not this_type.initialized:
            this_type.initialized = True

            self._logger = getAutomatonKitLogger()

            # If the landscape is in interactive mode, then all the coordinators should
            # default to using interactive mode
            self._interactive_mode = lscape.interactive_mode

            self._lscape_ref = weakref.ref(lscape)

            self._coord_lock = threading.RLock()

            self._cl_children = {}

            self._found_devices = []
            self._available_devices = []
            self._unavailable_devices = []

            self._coord_config = coord_config

            self._initialize(*args, **kwargs)
        return

    def _initialize(self, *_args, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization of the coordinator Singleton
            of a given type.
        """
        # pylint: disable=no-self-use
        raise AKitNotOverloadedError("_initialize: must be overloaded by derived coordinator classes") from None

    @property
    def available_devices(self):
        """
            The devices that the coordinator found to be available during startup.
        """
        return self._available_devices

    @property
    def children(self) -> List[LandscapeDevice]:
        """
            Returns a list of the devices created by the coordinator and registered by the coordinator with the Landscape object.
        """
        chlist = EMPTY_LIST

        self._coord_lock.acquire()
        try:
            chlist = [c.basedevice for c in self._cl_children.values()]
        finally:
            self._coord_lock.release()

        return chlist

    @property
    def children_as_extension(self) -> List[LandscapeDeviceExtension]:
        """
            Returns a list of the device protocol extensions created by this coordinator that have been attached to a landscape device.
        """
        chlist = EMPTY_LIST

        self._coord_lock.acquire()
        try:
            chlist = [c for c in self._cl_children.values()]
        finally:
            self._coord_lock.release()

        return chlist

    @property
    def coord_config(self):
        """
            The dedicated coordinator configuration for coordinators that have a dedicated
            configuration section in the landscape file.  Example: power, serial, wireless
        """

    @property
    def landscape(self):
        """
            Returns a hard reference to the Landscape singleton instance.
        """
        lscape = self._lscape_ref()
        return lscape

    @property
    def found_devices(self):
        """
            The devices that were dynamically discovery by the coordinators discovery protocol.
        """
        return self._found_devices

    @property
    def unavailable_devices(self):
        """
            The devices that the coordinator found to be unavailable during startup.
        """
        return self._unavailable_devices

    def establish_presence(self):
        """
            Implemented by derived coordinator classes to establish a specific presence in the
            landscape.
        """
        return

    def lookup_device_by_key(self, key) -> LandscapeDevice:
        """
            Looks up a device from the list of children by key in a thread safe way.
        """

        found = None

        self._coord_lock.acquire()
        try:
            if key in self._cl_children:
                found = self._cl_children[key].basedevice
        finally:
            self._coord_lock.release()

        return found

    def verify_connectivity(self, cmd: str = "echo 'It Works'", user: Optional[str] = None, raiseerror: bool = True):
        """
            Loops through the nodes in the coordinators pool in order to verify connectivity with the remote node.

            :param cmd: A command to run on the remote machine in order to verify that connectivity can be establish.
            :param user: The name of the user credentials to use for connectivity.
                         If the 'user' parameter is not provided, then the credentials
                         of the default or priviledged user will be used.
            :param raiseerror: A boolean value indicating if this API should raise an Exception on failure.

            :returns: A list of errors encountered when verifying connectivity with the devices managed or watched by the coordinator.
        """
        # pylint: disable=no-self-use
        raise AKitNotOverloadedError("verify_connectivity: must be overloaded by derived coordinator classes") from None



"""
.. module:: landscapebaselayer
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the Landscape related classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import threading

from akit.environment.context import Context

from akit.interop.landscaping.landscapedescription import LandscapeDescription
from akit.interop.landscaping.landscapedevice import LandscapeDevice
from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension
from akit.interop.landscaping.topologydescription import TopologyDescription

from akit.xlogging.foundations import getAutomatonKitLogger

class LandscapeBaseLayer:

    context = Context()

    logger = getAutomatonKitLogger()
    landscape_lock = threading.RLock()

    landscape_description = LandscapeDescription
    landscape_device = LandscapeDevice
    landscape_device_extension = LandscapeDeviceExtension

    topology_description = TopologyDescription



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ConnectionManager1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'ConnectionManager1' service.
    """

    SERVICE_MANUFACTURER = 'MicrosoftCorporation'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:ConnectionManager:1'

    SERVICE_EVENT_VARIABLES = {
        "CurrentConnectionIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "SinkProtocolInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "SourceProtocolInfo": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetCurrentConnectionIDs(self, extract_returns=True):
        """
            Calls the GetCurrentConnectionIDs action.

            :returns: "ConnectionIDs"
        """
        arguments = { }

        out_params = self._proxy_call_action("GetCurrentConnectionIDs", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ConnectionIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCurrentConnectionInfo(self, ConnectionID, extract_returns=True):
        """
            Calls the GetCurrentConnectionInfo action.

            :returns: "RcsID", "AVTransportID", "ProtocolInfo", "PeerConnectionManager", "PeerConnectionID", "Direction", "Status"
        """
        arguments = {
            "ConnectionID": ConnectionID,
        }

        out_params = self._proxy_call_action("GetCurrentConnectionInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RcsID", "AVTransportID", "ProtocolInfo", "PeerConnectionManager", "PeerConnectionID", "Direction", "Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetProtocolInfo(self, extract_returns=True):
        """
            Calls the GetProtocolInfo action.

            :returns: "Source", "Sink"
        """
        arguments = { }

        out_params = self._proxy_call_action("GetProtocolInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Source", "Sink",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: automationpodcoupling
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains a UpnpCoordinatorIntegration object to use for working with the nodes of a cluster

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Tuple, TYPE_CHECKING

from akit.environment.contextpaths import ContextPaths
from akit.exceptions import AKitConfigurationError, AKitSemanticError
from akit.coupling.coordinatorcoupling import CoordinatorCoupling
from akit.interop.coordinators.upnpcoordinator import UpnpCoordinator

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class UpnpCoordinatorIntegration(CoordinatorCoupling):
    """
        The UpnpCoordinatorIntegration handle the requirement registration for the UPNP coordinator.
    """

    pathbase = "/upnp"

    def __init__(self, *args, **kwargs):
        """
            The default contructor for an :class:`UpnpCoordinatorIntegration`.
        """
        super(UpnpCoordinatorIntegration, self).__init__(*args, **kwargs)
        return

    @classmethod
    def attach_to_environment(cls, constraints: Dict={}):
        """
            This API is called so that the IntegrationCoupling can process configuration information.  The :class:`IntegrationCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        resources_acquired = False

        upnp_device_hints = cls.landscape.get_upnp_device_config_lookup_table()
        if len(upnp_device_hints) > 0:
            resources_acquired = True

        if resources_acquired:
            cls.landscape.activate_integration_point("coordinator/upnp", cls.create_coordinator)
        else:
            errmsg = "The required UPNP resources were not configured."
            raise AKitConfigurationError(errmsg) from None

        return

    @classmethod
    def attach_to_framework(cls, landscape: "Landscape"):
        """
            This API is called so that the IntegrationCoupling can attach to the test framework and participate with
            registration processes.  This allows the framework to ignore the bringing-up of couplings that are not being
            included by a test.
        """
        super(UpnpCoordinatorIntegration, cls).attach_to_framework(landscape)
        cls.landscape.register_integration_point("coordinator/upnp", cls)
        return

    @classmethod
    def collect_resources(cls):
        """
            This API is called so the `IntegrationCoupling` can connect with a resource management
            system and gain access to the resources required for the automation run.

            :raises :class:`akit.exceptions.AKitResourceError`:
        """
        return

    @classmethod
    def create_coordinator(cls, landscape: "Landscape") -> object:
        """
            This API is called so that the landscape can create a coordinator for a given integration role.
        """
        cls.coordinator = UpnpCoordinator(landscape)
        return cls.coordinator

    @classmethod
    def declare_precedence(cls) -> int:
        """
            This API is called so that the IntegrationCoupling can declare an ordinal precedence that should be
            utilized for bringing up its integration state.
        """
        # We need to call the base class, it sets the 'logger' member
        super(UpnpCoordinatorIntegration, cls).declare_precedence()
        return

    @classmethod
    def diagnostic(cls, label: str, level: int, diag_folder: str):
        """
            The API is called by the :class:`akit.sequencer.Sequencer` object when the automation sequencer is
            building out a diagnostic package at a diagnostic point in the automation sequence.  Example diagnostic
            points are:

            * pre-run
            * post-run

            Each diagnostic package has its own storage location so derived :class:`akit.scope.ScopeCoupling` objects
            can simply write to their specified output folder.

            :param label: The label associated with this diagnostic.
            :param level: The maximum diagnostic level to run dianostics for.
            :param diag_folder: The output folder path where the diagnostic information should be written.
        """

        return

    @classmethod
    def establish_connectivity(cls, allow_missing_devices: bool=False, upnp_recording: bool = False, allow_unknown_devices: bool = False) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish connectivity with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        lscape = cls.landscape

        exclude_interfaces = ["lo"]

        runtime_exclude_interfaces = cls.context.lookup(ContextPaths.UPNP_EXCLUDE_INTERFACES)
        if runtime_exclude_interfaces is not None:
            exclude_interfaces = runtime_exclude_interfaces

        upnp_hint_table = lscape.get_upnp_device_config_lookup_table()
        
        if len(upnp_hint_table) == 0:
            raise AKitSemanticError("we should not have been called if the upnp device config had 0 devices.") from None

        required_devices = None
        if not allow_missing_devices:
            required_devices = [usn for usn in upnp_hint_table.keys()]

        found_device_results, matching_device_results, missing_device_results = cls.coordinator.startup_scan(
            upnp_hint_table, watchlist=upnp_hint_table, required_devices=required_devices, exclude_interfaces=exclude_interfaces,
            upnp_recording=upnp_recording, allow_unknown_devices=allow_unknown_devices)

        conn_results = {
            "upnp": {
                "found": found_device_results,
                "matching": matching_device_results,
                "missing": missing_device_results
            }
        }

        conn_errors = []

        return (conn_errors, conn_results)

    @classmethod
    def establish_presence(cls) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish presence with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        cls.coordinator.establish_presence()
        return

    def checkout_upnp_device(self, usn):
        """
            Checkout a device from the device pool by USN.
        """
        codev = None

        if usn in self.upnp_devices_pool:
            codev = self.upnp_devices_pool.pop(usn)
            self.upnp_devices_inuse[usn] = codev

        return codev

    def checkin_upnp_device(self, codev):
        """
            Checkin a device to the device pool.
        """
        usn = codev["USN"]

        if usn in self.upnp_devices_inuse:
            codev = self.upnp_devices_inuse.pop(usn)
            self.upnp_devices_pool[usn] = codev

        return

    



"""
.. module:: akit.interop.upnp.extensions.sonos.zps12
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the a Upnp device for a Sonos Zps12.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.extensible import LoadableExtension

from akit.interop.upnp.extensions.standard.rootdevices.sonos.sonosplayer import SonosPlayer
from akit.interop.upnp.devices.upnprootdevice import UpnpRootDevice

class SonosDeviceZpS12(SonosPlayer, LoadableExtension):
    """
    """

    MANUFACTURER = "Sonos, Inc."
    MODEL_NUMBER = "S12"
    MODEL_DESCRIPTION = "Sonos Play:1"



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional

import os

from functools import wraps

from akit.environment.variables import AKIT_VARIABLES

class WELLKNOWN_PORTALS:
    TEST_DISCOVERY = "test-discovery"
    TESTRUN_START = "testrun-start"

class NetherPortal:
    """
        When you open a portal, if you are not time traveling, then you must pass
        from one realm to the next by going through 'The Nether'. There isn't much
        going on in 'The Nether' but keep a sharp eye out for 'The Creeper'.
    """
    
    def __init__(self):
        return

    def __enter__(self):

        # ░░░░░░░░░░░░░░░░░░░░░░░░░░
        # ░▓▓▓▒▒▒▓▓▓███▒▒▒███▓▓▓▒▒▒░
        # ░███▓▓▓▒▒▒▓▓▓███▒▒▒███▓▓▓░
        # ░▓▓▓░░░░░░▒▒▒▓▓▓░░░░░░███░
        # ░▒▒▒░░░░░░▓▓▓███░░░░░░▓▓▓░
        # ░▓▓▓▒▒▒███░░░░░░▓▓▓▒▒▒███░
        # ░▒▒▒███░░░░░░░░░░░░███▓▓▓░
        # ░███▓▓▓░░░░░░░░░░░░▒▒▒███░
        # ░▒▒▒███░░░███▒▒▒░░░▓▓▓▒▒▒░
        # ░▓▓▓▒▒▒▓▓▓▒▒▒███▓▓▓▒▒▒███░
        # ░▓▓▓▒▒▒▓▓▓▒▒▒███▓▓▓▒▒▒███░
        # ░░░░░░░░░░░░░░░░░░░░░░░░░░
        #    ░▒▒▒▓▓▓███▓▓▓▒▒▒███░
        #    ░▓▓▓▒▒▒▓▓▓███▓▓▓▒▒▒░
        #    ░███▓▓▓▒▒▒▓▓▓███▓▓▓░
        #    ░▒▒▒███▓▓▓███▓▓▓▒▒▒░
        #    ░███▓▓▓▒▒▒▓▓▓▒▒▒███░
        #    ░▓▓▓▒▒▒███▒▒▒▓▓▓▒▒▒░
        #    ░▒▒▒▓▓▓▒▒▒▓▓▓▒▒▒▓▓▓░
        # ░░░░░░░░░░░░░░░░░░░░░░░░░░
        # ░███▓▓▓▓▒▒▒▒░░▓▓▓▓████▒▒▒░
        # ░▒▒▒████▓▓▓▓░░████▒▒▒▒▓▓▓░
        # ░███▓▓▓▓████░░▓▓▓▓████▒▒▒░
        # ░░░░░░░░░░░░░░░░░░░░░░░░░░

        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        return False


THE_NETHER = NetherPortal()

timetravel_session_name = None

class TimePortal:
    """
        By creating a time portal, you pass into a place where time is somewhat
        expanded, because your generating tons of data, but where every trace of
        your existance is mapped.
    """
    
    def __init__(self, session_name):
        self._session_name = session_name
        self._tracer = None
        return

    def __enter__(self):
        # pycrunch_trace has a difficult way to setup configuration, so
        # don't import the "Trace" object until we have setup the
        # output directory and are ready to load the config.
        from pycrunch_trace.client.api.trace import Trace

        self._tracer = Trace()
        self._tracer.start(self._session_name)
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        self._tracer.stop()
        return False


def timeportal(portal_name: str, *, activations: Optional[List[str]]=None):
    """
       Starts tracing using PyCrunch tracing toolkit
    """
    def _decorator(func):

        @wraps(func)
        def wrapper(*args, **kws):
        
            session_name = str(func.__name__)

            with open_time_portal(portal_name, activations=activations) as portal:
                results = func(*args, **kws)
                return results

        return wrapper

    return _decorator


def open_time_portal(portal_name: str, activations: List[str]=[]):
    """
    """

    global timetravel_session_name

    portal = THE_NETHER

    activate_time_portal = False
    if AKIT_VARIABLES.AKIT_TIMETRAVEL:
        if portal_name in AKIT_VARIABLES.AKIT_TIMEPORTALS:
            activate_time_portal = True
        if len(activations) > 0:
            for akey in activations:
                if akey in AKIT_VARIABLES.AKIT_TIMEPORTALS:
                    activate_time_portal = True

    if activate_time_portal:
        if timetravel_session_name is None:
            from pycrunch_trace.config import config as pytrace_config
            from pathlib import Path
            from akit.paths import get_path_for_output

            output_path = get_path_for_output()

            timetravel_session_name = "timetravel_" + AKIT_VARIABLES.AKIT_RUNID.replace("-", "_")

            timetravel_directory = os.path.join(output_path, "timetravel")
            if not os.path.exists(timetravel_directory):
                os.makedirs(timetravel_directory)

            session_directory = os.path.join(output_path, "timetravel", timetravel_session_name)
            if not os.path.exists(session_directory):
                os.makedirs(session_directory)

            pytrace_config.working_directory = Path(output_path)
            pytrace_config.recording_directory = Path(timetravel_directory)

            portal = TimePortal(timetravel_session_name)

    return portal

def timemachine_timeportal_code_append(timeportal_name: str, code_lines: list, current_indent: str):

    add_indent = ""

    if AKIT_VARIABLES.AKIT_TIMETRAVEL and timeportal_name in AKIT_VARIABLES.AKIT_TIMEPORTALS:
        code_lines.append('')
        code_lines.append('{}# The time-travel flag was set and the "{}" portal was activated.'.format(
                            current_indent, timeportal_name))
        code_lines.append('{}from akit.timemachine import open_time_portal'.format(current_indent))
        code_lines.append('{}with open_time_portal("{}") as portal:'.format(current_indent, timeportal_name))

        add_indent = "    "

    return add_indent


"""
.. module:: landscapedescription
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`LandscapeDescription` class.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, TYPE_CHECKING

import copy
import json
import os
import shutil
import traceback
import yaml

from akit.environment.contextpaths import ContextPaths

from akit.exceptions import AKitConfigurationError, AKitRuntimeError
from akit.environment.context import Context
from akit.xlogging.foundations import getAutomatonKitLogger

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class LandscapeDescription:
    """
        The base class for all derived :class:`LandscapeDescription` objects.  The
        :class:`LandscapeDescription` is used to load a description of the entities
        and resources in the tests landscape that will be used by the tests.
    """

    @classmethod
    def register_integration_points(cls, landscape: "Landscape"):
        """
            Method called during the test framework ininitalization in order to register integartion couplings and their
            associated roles with the test framework.

            :param landscape: A reference to the landscape singleton object.  We pass in the landscape parameter in order
                              to eliminate the need to import the landscape module which would cause a circular reference.
        """
        return

    def load(self, landscape_file: str, log_to_directory: Optional[str]=None):
        """
            Loads and validates the landscape description file.
        """
        logger = getAutomatonKitLogger()

        landscape_info = None

        with open(landscape_file, 'r') as lf:
            lfcontent = lf.read()
            landscape_info = yaml.safe_load(lfcontent)

        if log_to_directory is not None:
            try:
                landscape_file_basename = os.path.basename(landscape_file)
                landscape_file_basename, landscape_file_ext = os.path.splitext(landscape_file_basename)

                landscape_file_copy = os.path.join(log_to_directory, "landscape-declared{}".format(landscape_file_ext))
                shutil.copy2(landscape_file, landscape_file_copy)

                # Create a json copy of the landscape file until the time when we can
                # parse yaml in the test summary javascript.
                landscape_info_copy = copy.deepcopy(landscape_info)

                landscape_file_copy = os.path.join(log_to_directory, "landscape-declared.json")
                with open(landscape_file_copy, 'w') as lsf:
                    json.dump(landscape_info_copy, lsf, indent=4)
            except Exception as xcpt:
                err_msg = "Error while logging the landscape file (%s)%s%s" % (
                    landscape_file, os.linesep, traceback.format_exc())
                raise AKitRuntimeError(err_msg) from xcpt

        errors, warnings = self.validate_landscape(landscape_info)

        if len(errors) > 0:
            errmsg_lines = [
                "ERROR Landscape validation failures:"
            ]
            for err_path, err_msg in errors:
                errmsg_lines.append("    {}: {}".format(err_path, err_msg))

            errmsg = os.linesep.join(errmsg_lines)
            raise AKitConfigurationError(errmsg) from None

        if len(warnings) > 0:
            for wrn_path, wrn_msg in warnings:
                logger.warn("Landscape Configuration Warning: ({}) {}".format(wrn_path, wrn_msg))

        if "devices" in landscape_info["pod"]:
            devices = landscape_info["pod"]["devices"]

            device_lookup_table = {}
            for dev in devices:
                dev_type = dev["deviceType"]
                if dev_type == "network/upnp":
                    upnp_info = dev["upnp"]
                    dkey = None
                    if "hint" in upnp_info:
                        dkey = upnp_info["hint"]
                    elif "USN" in upnp_info:
                        dkey = upnp_info["USN"]
                    dkey = "UPNP:{}".format(dkey.upper())
                    device_lookup_table[dkey] = dev
                elif dev_type == "network/ssh":
                    dkey = "SSH:{}".format(dev["host"]).upper()
                    device_lookup_table[dkey] = dev

            ctx = Context()
            skipped_devices = ctx.lookup(ContextPaths.SKIPPED_DEVICES, default=[])

            for dev_key in skipped_devices:
                dev_key = dev_key.upper()
                if dev_key in device_lookup_table:
                    device = device_lookup_table[dev_key]
                    device["skip"] = True

        return landscape_info

    def validate_landscape(self, landscape_info):
        """
            Validates the landscape description file.
        """
        errors = []
        warnings = []

        if "pod" in landscape_info:
            podinfo = landscape_info["pod"]
            if "devices" in podinfo:
                devices_list = podinfo["devices"]
                child_errors, child_warnings = self.validate_devices_list(devices_list, prefix="")
                errors.extend(child_errors)
                warnings.extend(child_warnings)
            elif "environment" in podinfo:
                envinfo = landscape_info["environment"]
                child_errors, child_warnings = self.validate_environment(envinfo)
                errors.extend(child_errors)
                warnings.extend(child_warnings)
            else:
                errors.append(["/pod/devices", "A pod description requires a 'devices' list data member."])
        else:
            errors.append(["/pod", "A landscape description requires a 'pod' data member."])

        return errors, warnings

    def validate_devices_list(self, devlist, prefix=""): # pylint: disable=unused-argument
        """
            Verifies that all the devices in a device list are valid and returns a list of errors found.
        """
        errors = []
        warnings = []

        for devidx, devinfo in enumerate(devlist):
            item_prefix = "/devices[%d]" % devidx
            child_errors, child_warnings = self.validate_device_info(devinfo, prefix=item_prefix)
            errors.extend(child_errors)
            warnings.extend(child_warnings)

        return errors, warnings

    def validate_device_info(self, devinfo, prefix=""):
        """
            Verifies that a device info dictionary has the required common fields and also has valid
            information for the declared device type.  Returns a list of errors found.

            Required Common Fields:
                deviceType

            Valid Device Types:
                network/ssh
                network/upnp
        """
        errors = []
        warnings = []

        if "deviceType" in devinfo:
            deviceType = devinfo["deviceType"]
            if deviceType == "network/upnp":
                if "upnp" in devinfo:
                    upnpinfo = devinfo["upnp"]
                    child_errors, child_warnings = self.validate_upnp_info(upnpinfo, prefix=prefix + "/upnp")
                    errors.extend(child_errors)
                    warnings.extend(child_warnings)
                else:
                    errors.append(prefix + "upnp", "Device type 'network/upnp' must have a 'upnp' data member.")
            if deviceType == "network/ssh":
                if "host" not in devinfo:
                    errors.append("SSH Devices must have a 'host' field.")
                if "credentials" not in devinfo:
                    errors.append("Device type 'network/ssh' must have a 'credentials' data member.")
        else:
            errors.append(prefix + "deviceType", "Device information is missing the required 'deviceType' data member.")

        return errors, warnings

    def validate_environment(self, envinfo):
        """
        "environment":
            "label": "production"
        """
        errors = []
        warnings = []

        return errors, warnings

    def validate_upnp_info(self, upnpinfo, prefix=""): # pylint: disable=no-self-use,unused-argument
        """
            Verifies that a upnp info dictionary has valid data member combinations and can be used. Returns a
            list of errors found.
        """
        errors = []
        warnings = []

        if "hint" not in upnpinfo:
            if "USN" not in upnpinfo:
                errors.append((prefix + "USN", "UPnP information is missing a 'hint' or 'USN' data member."))
            else:
                warnings.append((prefix + "USN", "UPnP information 'USN' field has been replaced by the 'hint' field."))

        if "modelNumber" not in upnpinfo:
            errors.append((prefix + "modelNumber", "UPnP information is missing a 'modelNumber' data member."))
        if "modelName" not in upnpinfo:
            errors.append((prefix + "modelName", "UPnP information is missing a 'modelName' data member."))

        return errors, warnings



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANPOTSLinkConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANPOTSLinkConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANPOTSLinkConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "DataCompression": { "data_type": "string", "default": None, "allowed_list": None},
        "DataModulationSupported": { "data_type": "string", "default": None, "allowed_list": None},
        "DataProtocol": { "data_type": "string", "default": None, "allowed_list": None},
        "DelayBetweenRetries": { "data_type": "ui4", "default": None, "allowed_list": None},
        "Fclass": { "data_type": "string", "default": None, "allowed_list": None},
        "ISPInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "ISPPhoneNumber": { "data_type": "string", "default": None, "allowed_list": None},
        "LinkType": { "data_type": "string", "default": None, "allowed_list": "['PPP_Dialup']"},
        "NumberOfRetries": { "data_type": "ui4", "default": None, "allowed_list": None},
        "PlusVTRCommandSupported": { "data_type": "boolean", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetCallRetryInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCallRetryInfo action.

            :returns: "NewNumberOfRetries", "NewDelayBetweenRetries"
        """
        arguments = { }

        out_params = self.call_action("GetCallRetryInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewNumberOfRetries", "NewDelayBetweenRetries",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDataCompression(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDataCompression action.

            :returns: "NewDataCompression"
        """
        arguments = { }

        out_params = self.call_action("GetDataCompression", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDataCompression",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDataModulationSupported(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDataModulationSupported action.

            :returns: "NewDataModulationSupported"
        """
        arguments = { }

        out_params = self.call_action("GetDataModulationSupported", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDataModulationSupported",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDataProtocol(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDataProtocol action.

            :returns: "NewDataProtocol"
        """
        arguments = { }

        out_params = self.call_action("GetDataProtocol", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDataProtocol",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFclass(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFclass action.

            :returns: "NewFclass"
        """
        arguments = { }

        out_params = self.call_action("GetFclass", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewFclass",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetISPInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetISPInfo action.

            :returns: "NewISPPhoneNumber", "NewISPInfo", "NewLinkType"
        """
        arguments = { }

        out_params = self.call_action("GetISPInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewISPPhoneNumber", "NewISPInfo", "NewLinkType",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPlusVTRCommandSupported(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPlusVTRCommandSupported action.

            :returns: "NewPlusVTRCommandSupported"
        """
        arguments = { }

        out_params = self.call_action("GetPlusVTRCommandSupported", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewPlusVTRCommandSupported",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetCallRetryInfo(self, NewNumberOfRetries, NewDelayBetweenRetries, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetCallRetryInfo action.
        """
        arguments = {
            "NewNumberOfRetries": NewNumberOfRetries,
            "NewDelayBetweenRetries": NewDelayBetweenRetries,
        }

        self.call_action("SetCallRetryInfo", arguments=arguments, aspects=aspects)

        return

    def action_SetISPInfo(self, NewISPPhoneNumber, NewISPInfo, NewLinkType, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetISPInfo action.
        """
        arguments = {
            "NewISPPhoneNumber": NewISPPhoneNumber,
            "NewISPInfo": NewISPInfo,
            "NewLinkType": NewLinkType,
        }

        self.call_action("SetISPInfo", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class SystemProperties1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'SystemProperties1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:SystemProperties:1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "CustomerID": { "data_type": "string", "default": None, "allowed_list": None},
        "ThirdPartyHash": { "data_type": "string", "default": None, "allowed_list": None},
        "UpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "UpdateIDX": { "data_type": "ui4", "default": None, "allowed_list": None},
        "VoiceUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    def action_AddAccountX(self, AccountType, AccountID, AccountPassword, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddAccountX action.

            :returns: "AccountUDN"
        """
        arguments = {
            "AccountType": AccountType,
            "AccountID": AccountID,
            "AccountPassword": AccountPassword,
        }

        out_params = self.call_action("AddAccountX", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AccountUDN",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddOAuthAccountX(self, AccountType, AccountToken, AccountKey, OAuthDeviceID, AuthorizationCode, RedirectURI, UserIdHashCode, AccountTier, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddOAuthAccountX action.

            :returns: "AccountUDN", "AccountNickname"
        """
        arguments = {
            "AccountType": AccountType,
            "AccountToken": AccountToken,
            "AccountKey": AccountKey,
            "OAuthDeviceID": OAuthDeviceID,
            "AuthorizationCode": AuthorizationCode,
            "RedirectURI": RedirectURI,
            "UserIdHashCode": UserIdHashCode,
            "AccountTier": AccountTier,
        }

        out_params = self.call_action("AddOAuthAccountX", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AccountUDN", "AccountNickname",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DoPostUpdateTasks(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DoPostUpdateTasks action.
        """
        arguments = { }

        self.call_action("DoPostUpdateTasks", arguments=arguments, aspects=aspects)

        return

    def action_EditAccountMd(self, AccountType, AccountID, NewAccountMd, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EditAccountMd action.
        """
        arguments = {
            "AccountType": AccountType,
            "AccountID": AccountID,
            "NewAccountMd": NewAccountMd,
        }

        self.call_action("EditAccountMd", arguments=arguments, aspects=aspects)

        return

    def action_EditAccountPasswordX(self, AccountType, AccountID, NewAccountPassword, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EditAccountPasswordX action.
        """
        arguments = {
            "AccountType": AccountType,
            "AccountID": AccountID,
            "NewAccountPassword": NewAccountPassword,
        }

        self.call_action("EditAccountPasswordX", arguments=arguments, aspects=aspects)

        return

    def action_EnableRDM(self, RDMValue, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EnableRDM action.
        """
        arguments = {
            "RDMValue": RDMValue,
        }

        self.call_action("EnableRDM", arguments=arguments, aspects=aspects)

        return

    def action_GetRDM(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRDM action.

            :returns: "RDMValue"
        """
        arguments = { }

        out_params = self.call_action("GetRDM", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RDMValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetString(self, VariableName, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetString action.

            :returns: "StringValue"
        """
        arguments = {
            "VariableName": VariableName,
        }

        out_params = self.call_action("GetString", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StringValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetWebCode(self, AccountType, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetWebCode action.

            :returns: "WebCode"
        """
        arguments = {
            "AccountType": AccountType,
        }

        out_params = self.call_action("GetWebCode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("WebCode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ProvisionCredentialedTrialAccountX(self, AccountType, AccountID, AccountPassword, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ProvisionCredentialedTrialAccountX action.

            :returns: "IsExpired", "AccountUDN"
        """
        arguments = {
            "AccountType": AccountType,
            "AccountID": AccountID,
            "AccountPassword": AccountPassword,
        }

        out_params = self.call_action("ProvisionCredentialedTrialAccountX", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IsExpired", "AccountUDN",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RefreshAccountCredentialsX(self, AccountType, AccountUID, AccountToken, AccountKey, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RefreshAccountCredentialsX action.
        """
        arguments = {
            "AccountType": AccountType,
            "AccountUID": AccountUID,
            "AccountToken": AccountToken,
            "AccountKey": AccountKey,
        }

        self.call_action("RefreshAccountCredentialsX", arguments=arguments, aspects=aspects)

        return

    def action_Remove(self, VariableName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Remove action.
        """
        arguments = {
            "VariableName": VariableName,
        }

        self.call_action("Remove", arguments=arguments, aspects=aspects)

        return

    def action_RemoveAccount(self, AccountType, AccountID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveAccount action.
        """
        arguments = {
            "AccountType": AccountType,
            "AccountID": AccountID,
        }

        self.call_action("RemoveAccount", arguments=arguments, aspects=aspects)

        return

    def action_ReplaceAccountX(self, AccountUDN, NewAccountID, NewAccountPassword, AccountToken, AccountKey, OAuthDeviceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReplaceAccountX action.

            :returns: "NewAccountUDN"
        """
        arguments = {
            "AccountUDN": AccountUDN,
            "NewAccountID": NewAccountID,
            "NewAccountPassword": NewAccountPassword,
            "AccountToken": AccountToken,
            "AccountKey": AccountKey,
            "OAuthDeviceID": OAuthDeviceID,
        }

        out_params = self.call_action("ReplaceAccountX", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAccountUDN",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetThirdPartyCredentials(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResetThirdPartyCredentials action.
        """
        arguments = { }

        self.call_action("ResetThirdPartyCredentials", arguments=arguments, aspects=aspects)

        return

    def action_SetAccountNicknameX(self, AccountUDN, AccountNickname, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAccountNicknameX action.
        """
        arguments = {
            "AccountUDN": AccountUDN,
            "AccountNickname": AccountNickname,
        }

        self.call_action("SetAccountNicknameX", arguments=arguments, aspects=aspects)

        return

    def action_SetString(self, VariableName, StringValue, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetString action.
        """
        arguments = {
            "VariableName": VariableName,
            "StringValue": StringValue,
        }

        self.call_action("SetString", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Callable, Dict, List

import inspect


from akit.exceptions import AKitSemanticError, AKitNotImplementedError


def mark_categories(*, categories: List[str]):
    """
        Used to mark a test with category markers.
    """

    def decorator(testfunc: Callable) -> Callable:

        if not hasattr(testfunc, "_metadata_"):
            testfunc._metadata_ = {}

            testfunc._metadata_["categories"] = categories

        return testfunc

    return decorator

def mark_keywords(*, keywords: List[str]):
    """
        Used to mark a test with a list of keywords.
    """

    def decorator(testfunc: Callable) -> Callable:

        if not hasattr(testfunc, "_metadata_"):
            testfunc._metadata_ = {}

            testfunc._metadata_["keywords"] = keywords

        return testfunc

    return decorator


def mark_priority(*, priority: int):
    """
        Used to mark a test with a priority.
    """

    def decorator(testfunc: Callable) -> Callable:

        if not hasattr(testfunc, "_metadata_"):
            testfunc._metadata_ = {}

            testfunc._metadata_["priority"] = str(priority)

        return testfunc

    return decorator


def mark_descendent_categories(*, categories: List[str]):
    """
        Used to mark a module and all tests that reside under it in scope with a list of categories.

        :param categories: A list of categories to assign to descendent groups and tests
    """
    
    caller_frame = inspect.stack()[1]
    calling_module = inspect.getmodule(caller_frame[0])

    if not hasattr(calling_module, "_metadata_"):
        calling_module._metadata_ = {}

    calling_module._metadata_["categories"] = categories

    return


def mark_descendent_keywords(*, keywords: List[str]):
    """
        Used to mark a module and all tests that reside under it in scope with a list of keywords.

        :param keywords: A list of keywords to assign to descendent groups and tests
    """
    
    caller_frame = inspect.stack()[1]
    calling_module = inspect.getmodule(caller_frame[0])

    if not hasattr(calling_module, "_metadata_"):
        calling_module._metadata_ = {}

    calling_module._metadata_["keywords"] = keywords

    return


def mark_descendent_priority(*, priority: int):
    """
        Used to mark a module and all tests that reside under it in scope with a priority.

        :param priority: A priority to assign te descendent groups and tests.
    """

    caller_frame = inspect.stack()[1]
    calling_module = inspect.getmodule(caller_frame[0])

    if not hasattr(calling_module, "_metadata_"):
        calling_module._metadata_ = {}

    calling_module._metadata_["priority"] = str(priority)

    return


class MetaFilter:

    def __init__(self, include: bool, group: str):
        self._include = include
        self._group = group
        return

    @property
    def group(self):
        return self._group

    @property
    def include(self):
        return self._include

    def should_include(self, metadata: Dict[str, Any]) -> bool:
        errmsg = "MetaFilter.should_include must be implemented in derived classes."
        raise AKitNotImplementedError(errmsg)


class MetaFilterContains(MetaFilter):

    def __init__(self, include: bool, group: str, value:str):
        super().__init__(include, group)
        self._value = value
        return
    
    def should_include(self, metadata: Dict[str, Any]) -> bool:

        include = False

        if metadata is not None:
            if self._group in metadata:
                found_marker = metadata[self._group]
                if isinstance(found_marker, str):
                    if found_marker.find(self._value) > -1:
                        include = True
                elif isinstance(found_marker, list) or isinstance(found_marker, tuple):
                    if self._value in found_marker:
                        include = True

        # If this is an exclude based match, then negate the logic
        # of the match
        if not self._include:
            include = False if include else True

        return include

class MetaFilterEquals(MetaFilter):

    def __init__(self, include: bool, group: str, value:str):
        super().__init__(include, group)
        self._value = value
        return

    def should_include(self, metadata: Dict[str, Any]) -> bool:

        include = False

        if metadata is not None:
            if self._group in metadata:
                found_marker = metadata[self._group]
                if found_marker == self._value:
                    include = True

        # If this is an exclude based match, then negate the logic
        # of the match
        if not self._include:
            include = False if include else True

        return include

class MetaFilterNotEquals(MetaFilterEquals):

    def __init__(self, include: bool, group: str, value:str):
        super().__init__(include, group, value)
        return

    def should_include(self, metadata: Dict[str, Any]) -> bool:

        include = super().should_include(metadata)
        include = False if include else True

        return include


def parse_marker_expression(expression: str):

    include = True
    group = None
    marker = None

    working_expr = expression.strip()

    include_symbol = working_expr[0]
    if include_symbol == "+":
        include = True
    elif include_symbol == "-":
        include = False
    else:
        errmsg = "Marker expressions must begin with a + (include) or - (exclude) symbol. expression={}".format(expression)
        raise AKitSemanticError(errmsg)

    # Strip off the include or exclude
    working_expr = working_expr[1:].lstrip()

    errmsg = "Marker expressions must be in the form (+ or -)(group)/(marker)(= or ~)(value). expression={}".format(expression)

    operator = None
    # Equals
    if working_expr.find("==") > 0:
        operator = "=="
        working_expr.split(operator)
    # Contains
    elif working_expr.find("~=") > 0:
        operator = "~="
    # Not Equal
    elif working_expr.find("!=") > 0:
        operator = "!="
    else:
        raise AKitSemanticError(errmsg)

    marker_group = None
    marker_value = None

    next_parts = working_expr.split(operator)
    if len(next_parts) == 1:
        if operator in ["==", "~=", "!="]:
            errmsg = 'Expressions using the "==", "~=", "!=" operators must have a value component. expression={}'.format(expression)
            raise AKitSemanticError(errmsg)
        marker_group = next_parts[0]
    elif len(next_parts) == 2:
        marker_group, marker_value = next_parts
    else:
        raise AKitSemanticError(errmsg)

    if operator == "==":
        metafilter = MetaFilterEquals(include, marker_group, marker_value)
    elif operator == "~=":
        metafilter = MetaFilterContains(include, marker_group, marker_value)
    elif operator == "!=":
        metafilter = MetaFilterNotEquals(include, marker_group, marker_value)
    else:
        errmsg = "Uh Oh. Did someone partially implement a new operator. expression={}".format(expression)
        raise AKitSemanticError(errmsg)

    return metafilter


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Tuple

import contextlib
import datetime
import email
import os
import urllib
import socket

from http.server import SimpleHTTPRequestHandler
from http import HTTPStatus

from akit.networking.httpserverthreadpool import HttpServerThreadPool

class SimpleWebContentHandler(SimpleHTTPRequestHandler):

    index_pages = ["index.html", "index.htm"]

    def __init__(self, request: socket.socket, client_address: Tuple[str, int], server: HttpServerThreadPool, 
                 directory: Optional[str]=None, index_pages: Optional[List[str]]=None, **_kwargs) -> None:
        """
            ..note: Overide the constructor for BaseHTTPRequestHandler so we can absorb any extra kwargs.
        """
        SimpleHTTPRequestHandler.__init__(self, request, client_address, server, directory=directory)
        
        if index_pages is not None:
            self.index_pages = index_pages

        self._kwargs = _kwargs
        return
    
    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.

        """
        path = self.translate_path(self.path)
        f = None
        if os.path.isdir(path):
            parts = urllib.parse.urlsplit(self.path)
            if not parts.path.endswith('/'):
                # redirect browser - doing basically what apache does
                self.send_response(HTTPStatus.MOVED_PERMANENTLY)
                new_parts = (parts[0], parts[1], parts[2] + '/',
                             parts[3], parts[4])
                new_url = urllib.parse.urlunsplit(new_parts)
                self.send_header("Location", new_url)
                self.end_headers()
                return None
            for index in self.index_pages:
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                return self.list_directory(path)
        ctype = self.guess_type(path)
        # check for trailing "/" which should return 404. See Issue17324
        # The test for this was added in test_httpserver.py
        # However, some OS platforms accept a trailingSlash as a filename
        # See discussion on python-dev and Issue34711 regarding
        # parseing and rejection of filenames with a trailing slash
        if path.endswith("/"):
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None
        try:
            f = open(path, 'rb')
        except OSError:
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None

        try:
            fs = os.fstat(f.fileno())
            # Use browser cache if possible
            if ("If-Modified-Since" in self.headers
                    and "If-None-Match" not in self.headers):
                # compare If-Modified-Since and time of last file modification
                try:
                    ims = email.utils.parsedate_to_datetime(
                        self.headers["If-Modified-Since"])
                except (TypeError, IndexError, OverflowError, ValueError):
                    # ignore ill-formed values
                    pass
                else:
                    if ims.tzinfo is None:
                        # obsolete format with no timezone, cf.
                        # https://tools.ietf.org/html/rfc7231#section-7.1.1.1
                        ims = ims.replace(tzinfo=datetime.timezone.utc)
                    if ims.tzinfo is datetime.timezone.utc:
                        # compare to UTC datetime of last modification
                        last_modif = datetime.datetime.fromtimestamp(
                            fs.st_mtime, datetime.timezone.utc)
                        # remove microseconds, like in If-Modified-Since
                        last_modif = last_modif.replace(microsecond=0)

                        if last_modif <= ims:
                            self.send_response(HTTPStatus.NOT_MODIFIED)
                            self.end_headers()
                            f.close()
                            return None

            self.send_response(HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(fs[6]))
            self.send_header("Last-Modified",
                self.date_time_string(fs.st_mtime))
            self.end_headers()
            return f
        except:
            f.close()
            raise

class SimpleWebServer(HttpServerThreadPool):

    def __init__(self, address: Tuple[str, int], directory: str, protocol: str, **kwargs):
        directory = os.path.abspath(os.path.expanduser(os.path.expandvars(directory)))
        
        SimpleWebContentHandler.protocol_version = protocol
        kwargs["directory"] = directory

        HttpServerThreadPool.__init__(self, address, SimpleWebContentHandler, **kwargs)
        return

    def get_server_address(self):
        """
            Get the address of the server.

            ..note: Overloaded to ensure this method will proxy well to remote processes.
        """
        return self.server_address

    def server_bind(self):
        # suppress exception when protocol is IPv4
        with contextlib.suppress(Exception):
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
        return super().server_bind()

    def server_close(self):
        super().server_close()
        return

    def server_start(self):
        """
            Start the server and thread pool.

            ..note: Overloaded to ensure this method will proxy well to remote processes.
        """
        HttpServerThreadPool.server_start(self)
        return
    
    def server_stop(self):
        """
            Stop the server and thread pool.

            ..note: Overloaded to ensure this method will proxy well to remote processes.
        """
        HttpServerThreadPool.server_stop(self)
        return


"""
.. module:: scopecoupling
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`ScopeCoupling` class and associated reflection methods.
        The :class:`ScopeCoupling` derived classes can be used to provide setup and teardown of test
        automation scopes of execution for groups of tests.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict

import inspect
import weakref

from akit.coupling.basecoupling import BaseCoupling
from akit.environment.context import ContextUser

class ScopeCoupling(BaseCoupling):
    """
        The :class:`ScopeCoupling` object is the base object class that is used for scope declaration. :class:`ScopeCoupling`
        derived objects are used to create a hierarchy of scopes that are representative of the scopes of execution
        that are represented by the runtime environment.  These scopes of execution are used to establish the runtime
        contexts that task and test instantiations can be run inside of.  The scopes of a runtime environment are
        typically hierarchical in nature starting with the root object of a tree and build more complexed
        environments as they the hierarchy is ascended.

        The code of the :class:`ScopeCoupling` is divided into class level code and instance level code.  The hierarchy
        of the :class:`ScopeCoupling` derived objects are used by the :class:`akit.testing.unittest.testsequencer.Sequencer`
        object to determine the grouping and order of the tests or tasks to be run.  The class level code of the
        :class:`ScopeCoupling` is run by the :class:`akit.sequencer.Sequencer` object based on the class hierarchy to
        establish the order that scopes are entered and exited as the automation sequence is executed by the
        :class:`akit.testing.unittest.testsequencer.Sequencer` object.  The :class:`ScopeCoupling` class level code, is executed
        before any object that inherits from a :class:`ScopeCoupling` derived object is instantiated, so the state for the
        scope has been establish.

        The :class:`ScopeCoupling` instance level code is utilized to inter-operate with the state of the scope and also
        provides scope specific functionality.

        ..notes :
            A scope represents a predefined state that is reached by the execution of code.  The state represents a
            requirement that is needed to be met in order for a task to be able to run.

            Scopes have a name that is like a file system path /configuration

            Scopes can contain state and they are deposited into the context in a leaf just like other nodes.

    """

    TEMPLATE_SCOPES_PREFIX = "/scopes/%s"

    pathname = None

    descendants = {}
    test_references = {}

    def __init__(self):
        """
            The default contructor for an :class:`ScopeCoupling`.
        """
        if self.pathname is None:
            scope_type = type(self)
            scope_leaf = (scope_type.__module__ + "." + scope_type.__name__).replace(".", "/")
            self.pathname = self.TEMPLATE_SCOPES_PREFIX % scope_leaf

        # Create a weak reference to this scope in the global context and create a finalizer that
        # will remove the weak reference when the scope object is destroyed
        wref = weakref.ref(self)
        weakref.finalize(self, scope_finalize, self.context, self.pathname)

        self.context.insert(self.pathname, wref)
        return

    @classmethod
    def attach_to_environment(cls, constraints: Dict={}):
        """
            This API is called so that the ScopeCoupling can process configuration information.  The :class:`ScopeCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        return

    def scope_enter(self):
        """
            This API is called by the sequencer when a scope of state is being entered by an automation
            run.  The derived `ScopeCoupling` implementation should initialize the scope they are designed
            to manage and if initialization fails, they should raise a :class:`akit.exceptions.AKitScopeEntryError`
            error.

            :raises :class:`akit.exceptions.AKitScopeEntryError`:
        """
        return

    def scope_exit(self):
        """
            This API is called by the sequencer when an automation run is exiting a scope.  The derived
            ScopeCoupling implementation should use this opportunity to tear down the scope that it is
            managing.
        """
        return

class ScopeAperture:
    def __init__(self, scope_type):
        self._scope_type = scope_type
        return

    def __enter__(self):
        self._scope_type.scope_enter()
        return

    def __exit__(self, err_inst, err_type, err_tb):
        self._scope_type.scope_exit() 
        return False

class IteratorScopeCoupling(ContextUser):
    """
        The :class:`IteratorScopeCoupling` object is the base object class that is used for interator scope declaration.
        :class:`IteratorScopeCoupling` derived objects are used to insert a state iteration context into a test scope.
    """

    @classmethod
    def iteration_initialize(cls):
        """
            This API is overridden by derived iterator scope couplings and is called by the sequencer at the start
            of the use of a scope before the scope is entered for the first time.  It provides a hook for the
            iteration scope to setup the iteration state for the iteration scope.
        """
        return

    @classmethod
    def iteration_advance(cls, iterctx): # pylint: disable=unused-argument
        """
            The 'iteration_advance' API is overridden by derived iterator scope couplings and is called by the
            sequencer after the scope exits.  This class level hook method is used by the sequencer to advance
            the scope to the next iteration state.  The 'iteration_advance' API will return a 'True' result
            when the advancement of the iteration state was successful and the scope can be re-entered for
            execution.  The 'iteration_advance' API will return a 'False' when the advancement of the iteration
            state has reached the end of its iteration cycle and the scope should not be re-entered.
        """
        return

def inherits_from_scope_coupling(cls) -> bool:
    """
        Helper function that is used to determine if a type is an :class:`ScopeCoupling` subclass, but not
        the ScopeCoupling type itself.
    """
    is_scopemi = False
    if inspect.isclass(cls) and cls is not ScopeCoupling and issubclass(cls, ScopeCoupling):
        is_scopemi = True
    return is_scopemi

def is_iteration_scope_coupling(cls) -> bool:
    """
        Helper function that is used to determine if a type is an :class:`IteratorScopeCoupling` subclass, but not
        the :class:`IteratorScopeCoupling` type itself.
    """
    is_iterscopemi = False
    if inspect.isclass(cls) and cls is not ScopeCoupling and issubclass(cls, ScopeCoupling) and \
        hasattr(cls, "iteration_initialize") and hasattr(cls, "iteration_advance"):
        is_iterscopemi = True
    return is_iterscopemi

def scope_finalize(context, pathname):
    """
        Callback method used to finalize scope object and ensure they are unpublished from the
        global context.

        :param context: A reference to the context object.
        :param pathname: A string lookup path to the object in the context.
    """
    context.remove(pathname)
    return



from typing import Callable, Union, Type

from akit.testing.testplus.resourcelifespan import ResourceLifespan
from akit.testing.testplus.registration.sourcebase import SourceBase

class ResourceSource(SourceBase):

    def __init__(self, source_func: Callable, query_func: Union[Callable, None], resource_type: Type, constraints: dict):
        SourceBase.__init__(self, source_func, query_func, resource_type, constraints)
        return



"""
.. module:: clustercoupling
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains a ClusterIntegration object to use for working with the nodes of a cluster

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Tuple

from akit.coupling.integrationcoupling import IntegrationCoupling

class ClusterIntegration(IntegrationCoupling):
    """
        This is a mock playback device.
    """

    pathbase = "/clusters"

    def __init__(self, *args, role=None, **kwargs):
        """
            The default contructor for an :class:`IntegrationCoupling`.
        """
        super(ClusterIntegration, self).__init__(*args, role=role, **kwargs)

        if self.pathbase is None:
            raise ValueError("The 'pathbase' class member variable must be set to a unique name for each integration class type.")

        self.context.insert(self.pathbase , self)
        return

    @classmethod
    def attach_to_environment(cls, constaints: Dict={}):
        """
            This API is called so that the IntegrationCoupling can process configuration information.  The :class:`IntegrationCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        return

    @classmethod
    def collect_resources(cls):
        """
            This API is called so the `IntegrationCoupling` can connect with a resource management
            system and gain access to the resources required for the automation run.

            :raises :class:`akit.exceptions.AKitResourceError`:
        """

        return

    @classmethod
    def diagnostic(cls, diag_level: int, diag_folder: str):
        """
            The API is called by the :class:`akit.sequencer.Sequencer` object when the automation sequencer is
            building out a diagnostic package at a diagnostic point in the automation sequence.  Example diagnostic
            points are:

            * pre-run
            * post-run

            Each diagnostic package has its own storage location so derived :class:`akit.scope.ScopeCoupling` objects
            can simply write to their specified output folder.

            :param diag_level: The maximum diagnostic level to run dianostics for.
            :param diag_folder: The output folder path where the diagnostic information should be written.
        """

        return

    @classmethod
    def establish_connectivity(cls, allow_missing_devices: bool=False) -> Tuple[List[str], Dict]:
        """
            This API is called so that this subclass of the `IntegrationCoupling` can establish connectivity with any
            compute or storage resources.

            :raises :class:`akit.exceptins.AKitInitialConnectivityError`:
        """

        return

    @classmethod
    def establish_presence(cls) -> Tuple[List[str], Dict]:
        """
            This API is called so the `IntegrationCoupling` can establish presence with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        return


"""
.. module:: xinspect
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains additional helper functions for inspecting python code and objects

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
          indicate they extend the functionality of a base python module and the `x` is pre-pended to
          prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


import inspect
import sys

def get_current_function_name():
    """
        Gets the function name of the current function which is calling get_current_function_name.
    """
    caller_func_name = None

    caller_stack = inspect.stack()[1]
    caller_func_name = caller_stack.frame.f_code.co_name

    return caller_func_name

def get_caller_function_name():
    """
        Gets the function name of the calling function which is the parent of the function calling
        get_caller_function_name or returns None if there is not stack frame available.
    """
    caller_func_name = None

    try:
        caller_stack = inspect.stack()[2]
        caller_func_name = caller_stack.frame.f_code.co_name
    except IndexError:
        pass

    return caller_func_name

def get_module_ancestry(module_name: str):

    module_list = []

    module_parts = module_name.split(".")
    module_parts_len = len(module_parts)

    for i in range(1, module_parts_len):
        ancestor_parts = module_parts[:i]
        amname = ".".join(ancestor_parts)

        amod = sys.modules[amname]
        module_list.append(amod)
    
    amod = sys.modules[module_name]
    module_list.append(amod)

    return module_list


"""
.. module:: dnserrors
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the error types utilized by the multicast DNS code.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from akit.exceptions import AKitCommunicationsProtocolError

class DnsDecodeError(AKitCommunicationsProtocolError):
    """
        Error raised when the decoding of a DNS response fails.
    """
class DnsNonUniqueNameException(AKitCommunicationsProtocolError):
    """
        A non-unique name was encountered.
    """

class DnsNamePartTooLongException(AKitCommunicationsProtocolError):
    """
        The name part was too long
    """

class DnsBadTypeInNameException(AKitCommunicationsProtocolError):
    """
        Bad type in name
    """



"""
.. module:: looperqueue
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains a :class:`LooperQueue` which provides a thread-safe queue for
        for the :class:`Looper` and :class:`LooperPool`.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from threading import RLock, Semaphore

from akit.exceptions import AKitLooperError


class LooperQueueShutdown:
    """
        The :class:`LooperShutdown` object provides a mechanism to tell :class:`Looper` threads
        the queue is shutdown and to exit.
    """

class LooperQueue:
    """
        The :class:`LooperQueue` provides an encapsulation of a queue, semaphore, and lock combination
        to be utilized and passed as one object.  Makes it easy to share the queue, semaphore, and
        lock between the :class:`LooperPool` and the :class:`Looper` threaded objects.
    """

    def __init__(self):
        self._queue = []
        self._queue_available = Semaphore(value=0)
        self._queue_lock = RLock()
        self._queue_shutdown = None
        return

    def push_work(self, packet: object):
        """
            Pushes a work packet for the :class:`LooperPool` threads to work on.
        """
        available = 0

        self._queue_lock.acquire()
        try:
            if self._queue_shutdown is not None:
                raise AKitLooperError("The queue has been shutdown, no more work is allowed to be queued.") from None

            self._queue.append(packet)
            self._queue_available.release()

            available = len(self._queue)
        finally:
            self._queue_lock.release()

        return available

    def push_work_packets(self, packets: list):
        """
            Pushes a list of work packets for the :class:`LooperPool` threads to work on.
        """

        available = 0

        self._queue_lock.acquire()
        try:
            if self._queue_shutdown is not None:
                raise AKitLooperError("The queue has been shutdown, no more work is allowed to be queued.") from None

            self._queue.extend(packets)
            self._queue_available.release()

            available = len(self._queue)
        finally:
            self._queue_lock.release()

        return available

    def pop(self):
        """
            Remove the next work packet from the :class:`LooperQueue` work queue.
        """
        packet= None

        self._queue_available.acquire()
        self._queue_lock.acquire()
        try:
            if len(self._queue) > 0:
                packet = self._queue.pop(0)

                if self._queue_shutdown is not None:
                    self._queue_shutdown.release()
        finally:
            self._queue_lock.release()

        return packet

    def shutdown_and_wait(self, notices):
        """
            Starts the queue shutdown and waits for the last work time to be removed
            from the queue.
        """

        self._queue_lock.acquire()
        try:
            for _ in range(0, notices):
                self._queue.append(LooperQueueShutdown())

            wcount = (len(self._queue) - 1) * -1
            self._queue_shutdown = Semaphore(value=wcount)
        finally:
            self._queue_lock.release()

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, List

from enum import Enum

from akit.xfeature import FeatureMask, FeatureTag

class ConstraintKeys(str, Enum):
    REQUIRED_FEATURES = "required_features"
    EXCLUDED_FEATURES = "excluded_features"

class Constraints(FeatureMask):

    def __init__(self, *, required_features: List[FeatureTag]=None,
                       excluded_features: List[FeatureTag]=None,
                       checkout: bool=False, **kwargs):
        super().__init__(required_features=required_features,
                         excluded_features=excluded_features,
                         checkout=checkout, **kwargs)
        return

    def __call__(self, **kwargs: Any):
        inst = dict(self)
        inst.update(kwargs)
        return inst
    
    def __repr__(self):
        this_type = type(self)
        dict_repr = super().__repr__()
        repr_str = "{}(**{})".format(this_type.__name__, dict_repr)
        return repr_str

    def get_import_statement(self):
        this_type = type(self)
        import_statement = "from {} import {}".format(this_type.__module__, this_type.__name__)
        return import_statement

def merge_constraints(*args: Constraints) -> Constraints:
    """
        Takes a sequence of :class:`Contraints` objects and merges them into a single
        :class:`Constraints` object.

        ..note: In cases where this method is provided with constraints with overlapping
                values, the last constraint with a give value will overwrite any previous
                values.
    """

    combined: Constraints = Constraints()

    for av in args:
        combined.update(av)
    
    return combined



"""
.. module:: service
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that is utilized by interactive services to activate the environment
               with logging to rotating logs much like what a persistant service would need.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import tempfile

from logging.handlers import RotatingFileHandler

from akit.exceptions import AKitConfigurationError, AKitSemanticError
from akit.environment.variables import ActivationProfile, AKIT_VARIABLES

__activation_profile__ = ActivationProfile.Service

# Guard against attemps to activate more than one, activation profile.
if AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE is not None:
    errmsg = "An attempt was made to activate multiple environment activation profiles. profile={}".format(
        AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE
    )
    raise AKitSemanticError(errmsg)

AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE = ActivationProfile.Service

if "AKIT_SERVICE_NAME" not in os.environ:
    errmsg = "To use the AutomationKit to provide a service, you must " \
             "set the AKIT_SERVICE_NAME environment variable."
    raise AKitConfigurationError(errmsg)

service_name = os.environ["AKIT_SERVICE_NAME"]

AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = "INFO"
AKIT_VARIABLES.AKIT_SERVICE_NAME = service_name
AKIT_VARIABLES.AKIT_JOBTYPE = "service"
AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY = "~/akit/services/{}".format(service_name)

# For console activation we don't want to log to the console and we want
# to point the logs to a different output folder
os.environ["AKIT_LOG_LEVEL_CONSOLE"] = AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE
os.environ["AKIT_JOBTYPE"] = AKIT_VARIABLES.AKIT_JOBTYPE
os.environ["AKIT_OUTPUT_DIRECTORY"] = AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY

import akit.activation.base # pylint: disable=unused-import,wrong-import-position

from akit.xlogging.foundations import logging_initialize, LoggingDefaults # pylint: disable=wrong-import-position

LoggingDefaults.DefaultFileLoggingHandler = RotatingFileHandler
logging_initialize()



"""
.. module:: looper
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains a Looper for repeating a loop function useful
        for repeated, work packet or queue processing.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

from threading import Event, Thread

from akit.exceptions import AKitNotOverloadedError, AKitLooperError

from akit.xthreading.looperqueue import LooperQueue, LooperQueueShutdown

class Looper:
    """
        The :class:`Looper` is the worker thread for the :class:`LooperPool` object.
    """

    def __init__(self, queue:LooperQueue, name: Optional[str]=None, group: Optional[str]=None, daemon:Optional[bool]=None, **kwargs):
        self._name = name
        self._group = group
        self._kwargs = kwargs
        self._daemon = daemon

        self._running = False
        self._thread = None

        self._queue = queue

        self._exit_gate = None

        return

    def start(self):
        """
            Method for starting the looper.
        """
        self._exit_gate = Event()

        start_gate = Event()
        start_gate.clear()

        self._thread = Thread(target=self._loop_entry, name=self._name,
            args=(start_gate, self._queue), kwargs=None, daemon=self._daemon)

        self._thread.start()

        # Wait for the thread to start before we let the start method to return
        start_gate.wait()

        return

    def thread_get_name(self):
        return self._thread.name

    def thread_set_name(self, name):
        self._thread.name = name
        return

    def wait_for_exit(self, timeout: Optional[str]=None):
        """
            Method to wait for the looper thread to exit.
        """
        if self._exit_gate is None:
            raise AKitLooperError("Looper: wait_for_exit called before Looper was started.") from None

        self._exit_gate.wait(timeout=timeout)
        return

    def loop(self, packet) -> bool: # pylint: disable=no-self-use
        """
            Method that is overloaded by derived classes in order to implement a work loop.
        """
        raise AKitNotOverloadedError("Looper: _loop must be overloaded by derived classes.") from None

    def _loop_entry(self, start_gate: Event, queue: LooperQueue):
        """
            Protected method that serves as the thread target and that drives
            the work loop.
        """
        self._exit_gate.clear()

        self._running = True
        start_gate.set()
        start_gate = None

        while True:

            packet = queue.pop()

            # If we pop a LooperQueueShutdown from the queue
            # the queue is shut down and we should exit
            if isinstance(packet, LooperQueueShutdown):
                break

            # We may have been woken up so we can exit
            if packet is None or not self._running:
                break

            self.loop(packet)

            # end while True

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Tuple
from types import TracebackType

import telnetlib

from akit.aspects import AspectsCmd
from akit.compat import bytes_cast, str_cast

from akit.interfaces.icommandrunner import ICommandRunner

from akit.aspects import ActionPattern, AspectsCmd, LoggingPattern, DEFAULT_CMD_ASPECTS

class TcpSerialBase(ICommandRunner):

    def __init__(self, host: str, port: int=22, aspects: AspectsCmd=DEFAULT_CMD_ASPECTS):
        self._host = host
        self._port = port
        self._aspects = aspects
        return
    
    def _create_connection(self) -> telnetlib.Telnet:

        tnconn = telnetlib.Telnet(host=self._host, port=self._port)

        cmd_out = tnconn.read_until(self.NORMAL_PROMPT, 1)

        tnconn.write(b"\n")
        cmd_out = tnconn.read_until(self.NORMAL_PROMPT)

        tnconn.write(b"echo $PS1\n")
        ps1_out = tnconn.read_until(self.NORMAL_PROMPT)
        self.NORMAL_PROMPT = ps1_out.splitlines(False)[1]

        tnconn.write(b"export PS1=\"%s\"\n" % self.SERIAL_PROMPT)

        cmd_out = tnconn.read_until(self.SERIAL_PROMPT)
        cmd_out = tnconn.read_until(self.SERIAL_PROMPT)
        return tnconn

    def _restore_prompt(self, tnconn: telnetlib.Telnet):
        tnconn.write(b"echo $PS1\n")
        ps1_out = tnconn.read_until(self.NORMAL_PROMPT)
        return

class TcpSerialSession(TcpSerialBase):

    def __init__(self, host: str, port:int=22, session_basis: Optional["TcpSerialSession"]=None, aspects: AspectsCmd=DEFAULT_CMD_ASPECTS):
        super().__init__(host, port=port, aspects=aspects)
        self._session_basis = session_basis
        self._telnet_conn = None
        return

    def __enter__(self):
        """
            Method that implements the __enter__ symantics used for 'with' statements.
        """
        self._telnet_conn = self._create_connection()

        return self

    def __exit__(self, ex_type: type, ex_val: Exception, ex_tb: TracebackType) -> bool:
        """
            Method that implements the __exit__ symantics used for 'with' statements.

            :returns: Returns false indicating that exceptions are not handled.
        """
        self.close()
        handled = False
        return handled

    def close(self):
        """
            Closes the SSH session and the assocatied SSH connection.
        """
        # Only close the client if this session owns the client
        if self._session_basis is None:
            self._telnet_conn.close()
        return

    def open_session(self, session_basis: Optional["ICommandRunner"] = None,
                     aspects: Optional[AspectsCmd] = None) -> "ICommandRunner": # pylint: disable=arguments-differ
        """
            Provies a mechanism to create a :class:`TcpSerialSession` object with derived settings.  This method allows various
            parameters for the session to be overridden.  This allows for the performing of a series of TCP Serial operations under
            a particular set of shared settings or command behaviors.

            :param session_basis: An optional existing TcpSerialSession instance to use as a basis for the new session.
            :param aspects: The default run aspects to use for the operations performed by the session.
        """

        session_aspects = None
        if aspects is None:
            session_aspects = self._aspects

        session = None
        if session_basis is not None:
            session = TcpSerialSession(session_basis._host, port=session_basis._port, session_basis=session_basis, aspects=session_aspects)
        else:
            session = TcpSerialSession(self._host, port=self._port, aspects=session_aspects)

        return session

    
    def _create_connection(self) -> telnetlib.Telnet:
        """
            Create an Telnet connection if one is not already open and available.

            :returns: An :class:`telnetlib.Telnet` object connected to the remote Tcp Serial concentrator.
        """
        tnconn = None

        if self._session_basis is not None:
            tnconn = self._session_basis._telnet_conn
        else:
            tnconn = TcpSerialBase._create_connection(self)

        return tnconn

class TcpSerialAgent(TcpSerialBase):
    """
        The :class:`TcpSerialAgent` provides access to serial port hosted on a TCP/IP network.  The
        remote endpoint might be a host and port for a serial port hosted on a serial concentrator.
        It could be the host and port of a serial port hosted on a linux box via USB and ser2net.

        ..note:
            An example ser2net configuration might look like the configuration below.

            BANNER:banner:\r\nser2net port \p device \d [\s] (Debian GNU/Linux)\r\n\r\n

            3333:telnet:0:/dev/ttyUSB0:115200 8DATABITS NONE 1STOPBIT banner

            
    """

    SERIAL_PROMPT=b"@@@&@@@&"
    NORMAL_PROMPT=b"#"

    def __init__(self, host, port:int=22, aspects: AspectsCmd=DEFAULT_CMD_ASPECTS):
        super().__init__(host, port=port, aspects=aspects)
        return

    def open_session(self, session_basis: Optional["ICommandRunner"] = None,
                     aspects: Optional[AspectsCmd] = None) -> "ICommandRunner": # pylint: disable=arguments-differ
        """
            Provies a mechanism to create a :class:`TcpSerialSession` object with derived settings.  This method allows various
            parameters for the session to be overridden.  This allows for the performing of a series of TCP Serial operations under
            a particular set of shared settings or command behaviors.

            :param session_basis: An optional existing TcpSerialSession instance to use as a basis for the new session.
            :param aspects: The default run aspects to use for the operations performed by the session.
        """

        session_aspects = None
        if aspects is None:
            session_aspects = self._aspects

        session = None
        if session_basis is not None:
            session = TcpSerialSession(session_basis._host, port=session_basis._port, session_basis=session_basis, aspects=session_aspects)
        else:
            session = TcpSerialSession(self._host, port=self._port, aspects=session_aspects)

        return session

    def run_cmd(self, command: str, aspects: Optional[AspectsCmd] = None) -> Tuple[int, str, str]:
        """
            Runs a command on the designated host using the specified parameters.

            :param command: The command to run.
            :param aspects: The run aspects to use when running the command.

            :returns: The status, stderr and stdout from the command that was run.
        """
        status = None
        stdout = None
        stderr = None

        command = bytes_cast("%s 2> $TNSTDERR\n" % command)

        tnconn = self._create_connection()
        try:
            tnconn.write(b"export TNSTDERR=/tmp/tnstderr\n")
            cmdout = tnconn.read_until(self.SERIAL_PROMPT, timeout=5)

            tnconn.write(command)
            stdout_raw = tnconn.read_until(self.SERIAL_PROMPT, timeout=5)

            tnconn.write(b"echo $?\n")
            status_raw = tnconn.read_until(self.SERIAL_PROMPT, timeout=5)
            status_bytes = status_raw.splitlines(False)[1]

            tnconn.write(b"cat $TNSTDERR\n")
            stderr_raw = tnconn.read_until(self.SERIAL_PROMPT)

            stdout = "\n".join(str_cast(stdout_raw).splitlines(False)[1:-1])
            stderr = "\n".join(str_cast(stderr_raw).splitlines(False)[1:-1])
            status = int(status_bytes)
        finally:
            if tnconn is not None:
                self._restore_prompt(tnconn)
            tnconn.close()

        return status, stdout, stderr

    


if __name__ == "__main__":
    sagent = TcpSerialAgent("192.168.1.40", 3333)
    
    status, stdout, stderr = sagent.run_cmd("ls /jffs")

    print("STATUS:\n%s\n" % status)
    print("STDOUT:\n%s\n" % stdout)
    print("STDERR:\n%s\n" % stderr)


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Tuple, Type, Union

import os
import re

from akit.exceptions import AKitReaderError, AKitSemanticError

class RegExPattern:
    """
        The :class:`RegExPattern` object provides a way to combine a regular expression
        pattern with additional characteristic data that describe how the pattern should be
        used by the :class:`RegExReader` to match and process text content.
    """
    def __init__(self, pattern: Union[str, re.Pattern], *, match_type: str=None, destination: str=None,
                 required: bool=False, repeats: bool=False, strict: Optional[bool]=None, consume: bool=False):

        if match_type is not None and destination is None:
            errmsg = "RegExPattern: If 'match_type' is specified then you must provide a 'destination' parameter."
            raise AKitSemanticError(errmsg)

        if repeats and match_type is None:
            errmsg = "RegExPattern: If 'repeats=True', you must specify an 'match_type' parameter."
            raise AKitSemanticError(errmsg)

        self.pattern = pattern
        if isinstance(pattern, str):
            self.pattern = re.compile(pattern)
        self.match_type = match_type
        self.destination = destination
        self.required = required
        self.repeats = repeats
        self.strict = strict
        self.consume = consume
        return

    def __repr__(self) -> str:
        pattern_repr = 'RegExPattern(%r, match_type=%r, destination=%r, required=%r, repeats=%r, strict=%r)' % (
            self.pattern, self.match_type, self.destination, self.required, self.repeats, self.strict
        )
        return pattern_repr

    def __str__(self) -> str:
        pattern_str = 'pattern=%r match_type=%r destination=%r required=%r repeats=%r strict=%r' % (
            self.pattern, self.match_type, self.destination, self.required, self.repeats, self.strict
        )
        return pattern_str

class RegExMultiPattern:
    """
        The :class:`RegExMultiPattern` object provides a way to combine list regular expression
        patterns with a set of additional characteristic data that describe how the patterns should
        be used by the :class:`RegExReader` to match list of lines from text content.
    """
    def __init__(self, patterns: List[Union[str, re.Pattern]], *, match_type: str=None, destination: str=None,
                 required: bool=False, repeats: bool=False, strict: Optional[bool]=None, consume: bool=False):

        if match_type is not None and destination is None:
            errmsg = "RegExPattern: If 'match_type' is specified then you must provide a 'destination' parameter."
            raise AKitSemanticError(errmsg)
        
        self.destination = destination
        self.patterns = None

        if len(self.patterns) == 0:
            errmsg = "The RegExMultiPattern object requires at least one pattern to be passed."
            raise AKitSemanticError(errmsg)
        else:
            self.patterns = [p if isinstance(p, re.Pattern) else re.compile(p) for p in patterns]

        self.match_type = match_type
        self.required = required
        self.repeats = repeats
        self.strict = strict
        self.consume = consume
        return

    def __repr__(self) -> str:
        pattern_repr = 'RegExPattern(%r, match_type=%r, destination=%r, required=%r, repeats=%r, strict=%r)' % (
            self.patterns, self.match_type, self.destination, self.required, self.repeats, self.strict
        )
        return pattern_repr

    def __str__(self) -> str:
        pattern_str = 'patterns=%r match_type=%r destination=%r required=%r repeats=%r strict=%r' % (
            self.patterns, self.match_type, self.destination, self.required, self.repeats, self.strict
        )
        return pattern_str

class RegExReader:
    """
        The :class:`RegExReader` is a class that makes it easy to create text content
        parsers.
        
        The user of this class will inherit from the class and override the
        `EXPECTED_LINES` class variable and the `_process_matches` method to capture
        value matches.
        
        I The user might also want to override the `_register_pattern_match_bugs`
        method in order to integrate with custom enterprise but filing functionality
        for filing bugs for pattern match misses.
    """

    EXPECTED_LINES: List[Union[RegExPattern, RegExMultiPattern]] = None
    READ_ONLY_PATTERN = False
    CONSUME_WHITESPACE = True

    def __init__(self, content:Union[List, str], strict=False):
        self._content = content

        if isinstance(content, str):
            self._content = content.splitlines(keepends=False)

        self._strict = strict

        self._pattern_misses = []
        self._pattern_skips = []

        self._reader_code_start = 0

        self._process_content()
        return

    def _get_preview_line(self, content_queue):

        preview_line = None

        for nxt_line in content_queue:
            if self.CONSUME_WHITESPACE and nxt_line.strip() == "":
                continue
            preview_line = nxt_line

        return preview_line

    def _process_content(self):

        pattern_queue = [p for p in self.EXPECTED_LINES]
        content_queue = [cl for cl in self._content]

        if len(pattern_queue) == 0:
            errmsg = "The 'EXPECTED_LINES' list must be a list with at least one 'RegExPattern' or 'RegExMultiPattern' object."
            raise AKitSemanticError(errmsg)

        current_pattern = None
        current_line = None
        strict = None

        while len(content_queue) > 0:

            current_line = content_queue.pop(0)
            if self.CONSUME_WHITESPACE:
                while current_line.strip() == "" and len(content_queue) > 0:
                    current_line = content_queue.pop(0)

            if current_line is None:
                break

            # If pattern was set to None then we either found a match and
            # the pattern that was matched was not a repeating patter or
            # it was a repeating pattern but the repeating condition was
            # terminated due to mis-match or match for a following pattern.
            if current_pattern is None:
                if len(pattern_queue) == 0:
                    break

                current_pattern = pattern_queue.pop(0)

                strict = self._strict
                if current_pattern.strict is not None:
                    strict = current_pattern.strict

            matches = None
            if isinstance(current_pattern, RegExPattern):

                matches = self._process_pattern_simple(current_pattern, current_line, strict)
                if matches is not None:
                    if not current_pattern.consume:
                        if current_pattern.match_type is None:
                            self._process_name_based_match(current_pattern, matches)
                        else:
                            self._process_match_type_based_match(current_pattern, matches)

            elif isinstance(current_pattern, RegExMultiPattern):

                matches = self._process_pattern_multimatch(current_pattern, content_queue, strict)
                if matches is not None:
                    if not current_pattern.consume:
                        if current_pattern.match_type is None:
                            self._process_name_based_match(current_pattern, matches)
                        else:
                            self._process_match_type_based_match(current_pattern, matches)

            else:
                errmsg = "The 'EXPECTED_LINES' list must be a list of 'RegExPattern' or 'RegExMultiPattern' objects."
                raise AKitSemanticError(errmsg)

            if matches is not None:
                if not current_pattern.repeats:
                    current_pattern = None

                elif len(pattern_queue) > 0 and len(content_queue) > 0:
                    # If we are processing a repeating pattern, we need to see if the next
                    # pattern is going to match and result in a termination of the current
                    # repeating pattern.

                    preview_pattern = pattern_queue[0]
                    preview_line = self._get_preview_line(content_queue)

                    if preview_line is not None:
                        # Perform a preview operation, if the next pattern is a match for the
                        # next line then clear the current pattern
                        pmobj = None
                        if isinstance(preview_pattern, RegExPattern):
                            pmobj = preview_pattern.pattern.match(preview_line)
                        elif isinstance(preview_pattern, RegExMultiPattern):
                            pmobj = preview_pattern.patterns[0].match(preview_line)
                        if pmobj is not None:
                            current_pattern = None
                    else:
                        current_pattern = None

        if len(self._pattern_misses) > 0 or len(pattern_queue) > 0:
            # We shouldn't have pattern misses, if we have a miss, let the
            # reader register bugs.
            self._pattern_misses.extend(pattern_queue)
            self._register_pattern_match_bugs()

        return

    def _process_match_type_based_match(self, pattern: RegExPattern, matches: dict):

        destname = pattern.destination if not self.READ_ONLY_PATTERN else "_" + self.READ_ONLY_PATTERN

        pitem = pattern.match_type(**matches)
        if pattern.repeats:
            getattr(self, destname).append(pitem)
        else:
            setattr(self, destname, pitem)

        return

    def _process_name_based_match(self, pattern: RegExPattern, matches: dict):

        for mname, mval in matches.items():
            destname = mname if not self.READ_ONLY_PATTERN else "_" + mname
            setattr(self, destname, mval)

        return

    def _process_pattern_multimatch(self, multi_pattern: RegExMultiPattern, current_line: str, content_queue: str, strict: bool):

        match_complete = False
        current_matches = {}

        exp_pattern_list = [p for p in multi_pattern.patterns]

        cqindex = 0
        cqlen = len(content_queue)

        while len(exp_pattern_list) > 0:

            current_pattern = exp_pattern_list.pop(0)
            mobj = current_pattern.match(current_line)
            if mobj is not None:
                current_matches.update(mobj.groupdict())
            else:
                match_complete = False
                break

            if cqindex < cqlen:
                current_line = content_queue[cqindex]
                if self.CONSUME_WHITESPACE and current_line.strip() == "":
                    while current_line.strip() == "" and cqindex < (cqlen - 1):
                        cqindex += 1
                        current_line = content_queue[cqindex]
            else:
                break

            cqindex += 1

        matches = None

        if match_complete:
            matches = current_matches
            
            # If we did find a match, we need to consume the rest of the lines from the
            # queue that we used to make the match
            while cqindex > 0 and len(content_queue) > 0:
                content_queue.pop(0)
                cqindex -= 1

        elif not match_complete and strict:
            append_line_count = len(multi_pattern.patterns) - 1

            errmsg_lines = [
                "Failed to match content line with a strict pattern match.",
                "REGEX: %r" % multi_pattern.patterns,
                "LINES:",
                "    {}".format(current_line)
            ]

            for nxtindex, nxtline in enumerate(content_queue):
                if nxtindex >= append_line_count:
                    break
                errmsg_lines.append("    {}".format(nxtline))

            errmsg = os.linesep(errmsg_lines)
            raise AKitReaderError(errmsg)
        else:
            self._pattern_misses.append(multi_pattern)

        return matches

    def _process_pattern_simple(self, regex_pattern: RegExPattern, current_line: str, strict: bool):

        matches = None

        mobj = regex_pattern.pattern.match(current_line)
        if mobj is not None:
            if not regex_pattern.consume:
                matches = mobj.groupdict()
            else:
                matches = {}
        elif strict:
            errmsg_lines = [
                "Failed to match content line with a strict pattern match.",
                "REGEX: %r" % regex_pattern.pattern,
                "LINE: %r" % current_line 
            ]
            errmsg = os.linesep(errmsg_lines)
            raise AKitReaderError(errmsg)
        else:
            self._pattern_misses.append(regex_pattern)

        return matches

    def _register_pattern_match_bugs(self):

        return


"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Messaging2ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Messaging:2' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Messaging2'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "NewMessages": { "data_type": "string", "default": None, "allowed_list": None},
        "SessionUpdates": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AcceptSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AcceptSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("AcceptSession", arguments=arguments, aspects=aspects)

        return

    def action_CancelFileTransfer(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CancelFileTransfer action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("CancelFileTransfer", arguments=arguments, aspects=aspects)

        return

    def action_CloseSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CloseSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("CloseSession", arguments=arguments, aspects=aspects)

        return

    def action_CreateSession(self, SessionClass, SessionRecipients, Subject, SupportedContentType, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateSession action.

            :returns: "SessionID"
        """
        arguments = {
            "SessionClass": SessionClass,
            "SessionRecipients": SessionRecipients,
            "Subject": Subject,
            "SupportedContentType": SupportedContentType,
        }

        out_params = self.call_action("CreateSession", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SessionID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeleteMessage(self, MessageID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteMessage action.
        """
        arguments = {
            "MessageID": MessageID,
        }

        self.call_action("DeleteMessage", arguments=arguments, aspects=aspects)

        return

    def action_GetFileTransferSession(self, SessionID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFileTransferSession action.

            :returns: "FileInfoList"
        """
        arguments = {
            "SessionID": SessionID,
        }

        out_params = self.call_action("GetFileTransferSession", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FileInfoList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMessagingCapabilities(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMessagingCapabilities action.

            :returns: "SupportedCapabilities"
        """
        arguments = { }

        out_params = self.call_action("GetMessagingCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedCapabilities",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetNewMessages(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetNewMessages action.

            :returns: "NewMessages"
        """
        arguments = { }

        out_params = self.call_action("GetNewMessages", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewMessages",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSessionUpdates(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSessionUpdates action.

            :returns: "SessionUpdates"
        """
        arguments = { }

        out_params = self.call_action("GetSessionUpdates", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SessionUpdates",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSessions(self, SessionID, SessionClass, SessionStatus, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSessions action.

            :returns: "SessionsList"
        """
        arguments = {
            "SessionID": SessionID,
            "SessionClass": SessionClass,
            "SessionStatus": SessionStatus,
        }

        out_params = self.call_action("GetSessions", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SessionsList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTelephonyIdentity(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTelephonyIdentity action.

            :returns: "TelephonyIdentity"
        """
        arguments = { }

        out_params = self.call_action("GetTelephonyIdentity", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TelephonyIdentity",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_JoinSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the JoinSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("JoinSession", arguments=arguments, aspects=aspects)

        return

    def action_LeaveSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the LeaveSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("LeaveSession", arguments=arguments, aspects=aspects)

        return

    def action_ModifySession(self, SessionID, SessionRecipientsToAdd, SessionRecipientsToRemove, Subject, SupportedContentType, SessionClass, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ModifySession action.
        """
        arguments = {
            "SessionID": SessionID,
            "SessionRecipientsToAdd": SessionRecipientsToAdd,
            "SessionRecipientsToRemove": SessionRecipientsToRemove,
            "Subject": Subject,
            "SupportedContentType": SupportedContentType,
            "SessionClass": SessionClass,
        }

        self.call_action("ModifySession", arguments=arguments, aspects=aspects)

        return

    def action_ReadMessage(self, MessageID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReadMessage action.

            :returns: "MessageRequested"
        """
        arguments = {
            "MessageID": MessageID,
        }

        out_params = self.call_action("ReadMessage", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MessageRequested",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SearchMessages(self, MessageClass, MessageFolder, MessageStatus, SessionID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SearchMessages action.

            :returns: "MessageList"
        """
        arguments = {
            "MessageClass": MessageClass,
            "MessageFolder": MessageFolder,
            "MessageStatus": MessageStatus,
            "SessionID": SessionID,
        }

        out_params = self.call_action("SearchMessages", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MessageList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SendMessage(self, MessageToSend, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SendMessage action.

            :returns: "MessageID"
        """
        arguments = {
            "MessageToSend": MessageToSend,
        }

        out_params = self.call_action("SendMessage", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MessageID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_StartFileTransfer(self, FileInfoList, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartFileTransfer action.
        """
        arguments = {
            "FileInfoList": FileInfoList,
        }

        self.call_action("StartFileTransfer", arguments=arguments, aspects=aspects)

        return
